"""Algorithms for sound field control, in particular sound zone control. 


References
----------
[brunnstromSound2023] J. Brunnström, T. van Waterschoot, and M. Moonen, “Sound zone control for arbitrary sound field reproduction methods,” in European Signal Processing Conference (EUSIPCO), Helsinki, Finland, Sep. 2023. `[link] <https://doi.org/10.23919/EUSIPCO58844.2023.10289995>`__ \n
[brunnstromSignaltointerferenceplusnoise2023] J. Brunnström, T. van Waterschoot, and M. Moonen, “Signal-to-interference-plus-noise ratio based optimization for sound zone control,” IEEE Open Journal of Signal Processing, vol. 4, pp. 257–266, 2023, doi: 10.1109/OJSP.2023.3246398. `[link] <https://doi.org/10.1109/OJSP.2023.3246398>`__ \n
[leeFast2020] T. Lee, L. Shi, J. K. Nielsen, and M. G. Christensen, “Fast generation of sound zones using variable span trade-off filters in the DFT-domain,” IEEE/ACM Transactions on Audio, Speech, and Language Processing, vol. 29, pp. 363–378, Dec. 2020, doi: 10.1109/TASLP.2020.3042701. `[link] <https://doi.org/10.1109/TASLP.2020.3042701>`__ \n 


"""
import numpy as np
import copy
import scipy.linalg as splin

import aspcore.correlation as cr
import aspcore.matrices as mat
import aspcore.filter as fc
import aspcore.filterdesign as fd
import aspcore.fouriertransform as ft

# ================== UTILITIES ==================================
def beamformer_vec_to_ir(bf_vec, num_ls):
    """Transforms a beamformer vector into a shape more appropriate for a filter impulse response

    Parameters
    ----------
    bf_vec : ndarray of shape (num_ls*filt_len, 1) or (num_zones, num_ls*filt_len)

    Returns
    -------
    ir : ndarray of shape (1, num_ls, filt_len) or (num_zones, num_ls, filt_len)
    """
    bf_vec = np.squeeze(bf_vec)
    if bf_vec.ndim == 1:
        return bf_vec.reshape(num_ls, -1)
    elif bf_vec.ndim == 2:
        num_zones = bf_vec.shape[0]
        return bf_vec.reshape(num_zones, num_ls, -1)
    else:
        raise ValueError("Wrong dimensions for bf_vec")


def freq_to_time_beamformer(w, num_freqs):
    """Tranforms a frequency domain beamformer / control filter to a time domain impulse response
    
    Parameters
    ----------
    w : complex ndarray of shape ()
    num_freqs : int

    Returns
    -------
    ir : ndarray of shape ()
    """
    w = ft.insert_negative_frequencies(w, True)
    ir,_ = fd.fir_from_freqs_window(w, num_freqs-1)
    ir = np.moveaxis(ir, 0,1)
    ir = ir[:1,:,:]
    return ir



def spatial_cov_freq_superpos(Hb, Hd, d=None):
    """Calculates spatial covariance matrices from transfer function matrices

    Assumes a superposition model, so the transfer functions are bright zone and dark zone
    respecively.

    Parameters
    ----------
    Hb : ndarray of shape (num_freq, num_mic_bright, num_ls)
        complex transfer functions from loudspeakers to bright zone
    Hb : ndarray of shape (num_freq, num_mic_dark, num_ls)
        complex transfer functions from loudspeakers to dark zone
    d : ndarray of shape (num_freq, num_mic_bright, num_virt_src), optional
        desired complex sound pressure in the bright zone 

    Returns
    -------
    Rb : ndarray of shape (num_freq, num_ls, num_ls)
        Hermitian spatial covariance for the bright zone
    Rd : ndarray of shape (num_freq, num_ls, num_ls)
        Hermitian spatial covariance for the dark zone
    rb : ndarray of shape (num_freq, num_ls, num_virt_src)
        spatial cross-correlation between the desired pressure and the 
        transfer functions between loudspeakers and bright zone. Only 
        returned if d is provided.
    """
    assert Hb.ndim == 3 and Hd.ndim == 3
    assert Hb.shape[0] == Hd.shape[0]
    assert Hb.shape[2] == Hd.shape[2]
    num_micb = Hb.shape[1]
    num_micd = Hd.shape[1]
    Rb = np.moveaxis(Hb.conj(),1,2) @ Hb / num_micb
    Rd = np.moveaxis(Hd.conj(),1,2) @ Hd / num_micd
    
    if d is not None:
        assert d.ndim == 3
        assert d.shape[:2] == (Hb.shape[0], Hb.shape[1])
        rb = np.moveaxis(Hb.conj(),1,2) @ d / num_micb
        return Rb, Rd, rb
    else:
        return Rb, Rd

def fpaths_to_spatial_cov(arrays, fpaths, source_name, zone_names):
    """utility function to be used with aspsim package. Deprecated, and will be removed in future versions.

    - arrays is ArrayCollection object
    - fpaths is the frequency domain RIRs, see function get_fpaths() in this module
    - source name is a string for a source in arrays
    - zone_names is a list of strings to microphones in arrays

    returns a spatial covariance matrix R of shape (num_freqs, num_zones, num_src, num_src)
    """
    num_sources = arrays[source_name].num
    num_freqs = fpaths[source_name][zone_names[0]].shape[0]
    num_zones = len(zone_names)

    H = [fpaths[source_name][zone_name] for zone_name in zone_names]
    R = np.zeros((num_freqs, num_zones, num_sources, num_sources), dtype=complex)
    for k in range(num_zones):
        for f in range(num_freqs):
            R[f,k,:,:] = mat.ensure_pos_semidef(H[k][f,:,:].T.conj() @ H[k][f,:,:])
        num_mics = H[k][f].shape[0]
        R[:,k,:,:] /= num_mics
    return R

def paths_to_spatial_cov(arrays, source_name, zone_names, sources, filt_len, num_samples, margin=None):
    """utility function to be used with aspsim package. Deprecated, and will be removed in future versions.

    sources should be a list of the audio sources associated with each zone
        naturally the list of zone names and sources should be of the same length

    by default it will use as many samples as possible (only remove rir_len-1 samples 
        in the beginning since they haven't had time to propagate properly). 
        margin can be supplied if a specific number of samples should be removed instead.
        might give questionable result if you set margin to less than rir_len-1.
    

    Returns K^2 spatial covariance matrices R_{ki}, where k is the zones index of 
        the microphones and i is the zone index of the audio signal
        The returned array has shape (num_zones, num_zones, num_ls*ir_len, num_ls*ir_len)
        and is indexed as R_{ki} = R[k,i,:,:]
    """
    num_sources = arrays[source_name].num
    num_zones = len(zone_names)
    assert len(sources) == num_zones

    R = np.zeros((num_zones, num_zones, filt_len*num_sources, filt_len*num_sources), dtype=float)
    for k in range(num_zones):
        for i in range(num_zones):
            R[k,i,:,:] = mat.ensure_pos_semidef(spatial_cov(arrays.paths[source_name][zone_names[k]], sources[i], filt_len, num_samples, margin=margin))
    return R

def paths_to_spatial_cov_delta(arrays, source_name, zone_names, filt_len):
    """utility function to be used with aspsim package. Deprecated, and will be removed in future versions.

    See info for paths_to_spatial_cov
    """
    num_sources = arrays[source_name].num
    num_zones = len(zone_names)

    R = np.zeros((num_zones, num_zones, filt_len*num_sources, filt_len*num_sources), dtype=float)
    for k in range(num_zones):
        for i in range(num_zones):
            R[k,i,:,:] = mat.ensure_pos_semidef(spatial_cov_delta(arrays.paths[source_name][zone_names[k]], filt_len))
    return R

def rir_to_szc_cov(rir, ctrlfilt_len):
    """
    Takes a RIR of shape (num_ls, num_mic, ir_len) and
    turns it into the time domain sound zone control spatial
    covariance matrix made up of the blocks R_l1l2 = H_l1^T H_l2, 
    where H_l is a convolution matrix with RIRs associated with
    loudspeaker l

    output is of shape (num_ls*ctrlfilt_len, num_ls*ctrlfilt_len)
    """
    num_ls = rir.shape[0]
    num_mics = rir.shape[1]
    R = np.zeros((ctrlfilt_len*num_ls, ctrlfilt_len*num_ls))
    for m in range(num_mics):
        for l1 in range(num_ls):
            for l2 in range(num_ls):
                h1 = rir[l1,m,:]
                h2 = rir[l2,m,:]
                H1 = splin.convolution_matrix(h1, ctrlfilt_len ,mode="full")
                H2 = splin.convolution_matrix(h2, ctrlfilt_len ,mode="full")
                R[l1*ctrlfilt_len:(l1+1)*ctrlfilt_len, 
                    l2*ctrlfilt_len:(l2+1)*ctrlfilt_len] += H1.T @ H2
    R /= num_mics
    return R


def spatial_cov(ir, source, filt_len, num_samples, margin=None):
    """Computes time-domain spatial covariance matrix from room impulse responses

    The spatial covariance matrix is 
    $R = \\begin{bmatrix} = $

    Parameters
    ----------
    ir : ndarray of shape (num_ls, num_mic, ir_len)
        room impulse responses from loudspeakers to microphones
    source : Source object
        source object that generates the audio that should be reproduced in the sound zones. 
        The only requirement is that it is an object with a get_samples(num_samples) method, returning a 
        ndarray of shape (num_ls, num_samples).
    filt_len : int
        The length of the desired filter impulse response
    num_samples : int
        The number of samples to use for the spatial covariance matrix. Higher will give a more accurate estimate,
        but take longer to compute. 
    margin : int, optional
        by default the function will use as many samples as possible, which means only removing rir_len-1 samples 
        in the beginning of the filtered source signal, since those samples haven't had time to propagate properly. 
        margin can be supplied if a specific number of samples should be removed instead.
        might give questionable result if you set margin to less than rir_len-1.

    Returns
    -------
    R : ndarray of shape (num_ls*filt_len, num_ls*filt_len)
        The spatial covariance matrix
    
    Notes
    -----
    \\begin{equation}
        \\bm{R}_{zi} = \\expect \\bigl[ \\mathbb{X}_i^\\top (n) \\bm{H}_z \\bm{H}_z^{\\top} \\mathbb{X}_i(n)\\bigr] \\in \\mathbb{R}^{LI\\times LI}
    \\end{equation}

    """
    ir_len = ir.shape[-1]
    num_sources = ir.shape[0]
    if margin is None:
        margin = ir_len - 1

    rir_filt = fc.create_filter(ir=ir, sum_over_input=False)
    source = copy.deepcopy(source)
    in_sig = source.get_samples(num_samples+margin)
    in_sig = np.tile(in_sig, (num_sources, 1))
    out_sig = rir_filt.process(in_sig)
    out_sig = out_sig[...,margin:]
    out_sig = np.moveaxis(out_sig, 0, 1)
    R = cr.corr_matrix(out_sig, out_sig, filt_len, filt_len)
    return R


def spatial_cov_delta(ir, filt_len):
    """
    Calculates the spatial covariance matrices as if the input signal is a delta
    ir is the default shape given by arrays.paths (num_ls, num_mics, ir_len)
    
    Multiplies result with ir_len, because corr_matrix divides with the number of samples,
        but here that shouldn't be done to keep the correct scaling, as the values are just
        the ir and not filtered samples
        
    The returned spatial covariance matrix is of size (num_ls*filt_len, num_ls*filt_len)
    """
    ir_len = ir.shape[-1]
    ir = np.moveaxis(ir, 1, 0)
    R = cr.corr_matrix(ir, ir, filt_len, filt_len) * ir_len
    return R



# ================= CONVENTIONAL SOUND ZONE CONTROL ================
def acc(Rb, Rd, reg=0):
    """Frequency-domain acoustic contrast control for sound zone control

    Will calculate the principal generalized eigenvector of (Rb, Rd+reg*I)

    Parameters
    ----------
    Rb : ndarray of shape (num_freq, num_ls, num_ls)
        spatial covariance associated with the bright zone
    Rd : ndarray of shape (num_freq, num_ls, num_ls)
        spatial covariance associated with the dark zones
    reg : float
        non-negative, applies l2 regularization to the loudspeaker effort, by adding
        reg * np.eye() to the dark zone spatial covariance. 

    Returns
    -------
    w : ndarray of shape (num_freq, num_ls)
        the control filter used to generate loudspeaker signals.
    """
    assert Rb.shape == Rd.shape
    assert Rb.shape[-1] == Rb.shape[-2]
    assert Rb.ndim == 3
    num_freq = Rd.shape[0]
    num_ls = Rd.shape[-1]
    
    if reg > 0:
        Rd_reg = Rd + reg*np.eye(num_ls)[None,:,:]
    else:
        Rd_reg = Rd
    
    w = np.zeros((num_freq, num_ls))
    for f in range(num_freq):
        eigvals, evec = splin.eigh(Rb[f,:,:], mat.ensure_pos_def_adhoc(Rd_reg[f,:,:], verbose=True))
        w[f,:] = evec[:,-1] * np.sqrt(eigvals[-1])
    return w

def pressure_matching_szc(Rb, Rd, rb, mu, reg=0):
    """Frequency-domain pressure matching for sound zone control
    
    Parameters
    ----------
    Rb : ndarray of shape (num_freq, num_ls, num_ls)
        spatial covariance associated with the bright zone
    Rd : ndarray of shape (num_freq, num_ls, num_ls)
        spatial covariance associated with the dark zones
    rb : ndarray of shape (num_freq, num_ls, num_virt_src)
        cross correlation between paths to bright zone and virtual source paths
    mu : int
        non-negative, weights effort between error in dark and bright zones
    reg : float
        non-negative, applies l2 regularization to the loudspeaker effort, by adding
        reg * np.eye() to the dark zone spatial covariance

    Returns
    -------
    w : ndarray of shape (num_freq, num_ls, num_virt_src)
        the control filter used to generate loudspeaker signals.
    """
    assert Rb.shape == Rd.shape
    assert Rb.shape[-1] == Rb.shape[-2]
    assert Rb.ndim == 3
    num_freq = Rd.shape[0]
    num_ls = Rd.shape[-1]
    if reg > 0:
        Rd_reg = Rd + reg*np.eye(num_ls)[None,:,:]
    else:
        Rd_reg = Rd

    w = np.linalg.solve(Rb + mu*Rd_reg, rb)
    return w

def vast(Rb, Rd, rb, mu, rank, reg=0):
    """Frequency domain variable span trade-off filter for sound zone control 

    Parameters
    ----------
    Rb : ndarray of shape (num_freq, num_ls, num_ls)
        spatial covariance associated with the bright zone
    Rd : ndarray of shape (num_freq, num_ls, num_ls)
        spatial covariance associated with the dark zones
    rb : ndarray of shape (num_freq, num_ls, num_virt_src)
        cross correlation between paths to bright zone and virtual source paths
    mu : int 
        non-negative, weights effort between error in dark and bright zones
    rank : int
        between 1 and num_ls. Applies low-rank approximation via GEVD. 
        Lower rank gives higher acoustic contrast, higher rank gives lower distortion
    reg : float
        non-negative, applies l2 regularization to the loudspeaker effort, by adding
        reg * np.eye() to the dark zone spatial covariance

    Returns
    -------
    w : ndarray of shape (num_freq, num_ls, num_virt_src)
        the control filter used to generate loudspeaker signals. 
        loudspeaker signals can be generated from the virtual source signals 
        v : (num_freq, num_virt_src, 1) as w @ v. 

    References
    ----------
    [leeFast2020] T. Lee, L. Shi, J. K. Nielsen, and M. G. Christensen, “Fast generation of sound zones using variable span trade-off filters in the DFT-domain,” IEEE/ACM Transactions on Audio, Speech, and Language Processing, vol. 29, pp. 363–378, Dec. 2020, doi: 10.1109/TASLP.2020.3042701.
    """
    assert Rb.shape == Rd.shape
    assert Rb.shape[-1] == Rb.shape[-2]
    assert Rb.ndim == 3
    assert Rb.shape[0:2] == rb.shape[0:2]
    assert rb.ndim == 3
    num_freq = Rb.shape[0]
    num_ls = Rb.shape[-1]

    if reg > 0:
        Rd_reg = Rd + reg*np.eye(num_ls)[None,:,:]
    else:
        Rd_reg = Rd

    eigval = []
    eigvec = []
    for f in range(num_freq):
        eva, eve = splin.eigh(Rb[f,:,:], Rd_reg[f,:,:], type=1)
        eigval.append(eva[None,:])
        eigvec.append(eve[None,:,:])
    eigval = np.concatenate(eigval, axis=0)
    eigvec = np.concatenate(eigvec, axis=0)
    eigval = eigval[:,-rank:]
    eigvec = eigvec[:,:,-rank:]
    diag_entries = 1 / (eigval + mu*np.ones(rank))
    diag_mat = np.concatenate([np.diag(diag_entries[i,:])[None,:,:] for i in range(num_freq)], axis=0)
    w = eigvec @ diag_mat @ np.moveaxis(eigvec.conj(),1,2) @ rb
    return w




def acc_time(Rb, Rd, reg_param=0):
    """Time domain acoustic contrast control
    
    Will calculate the principal generalized eigenvector of (Rb, Rd+reg*I)

    OLD DOCUMENTATION:
    cov_bright and cov_dark is the szc spatial covariance matrices
    = H.T @ H, where H is a convolution matrix made up of the RIR, 
    summed over the appropriate microphones for bright and dark zones. 

    num_ls is the number of loudspeakers, and therefore the number 
    of blocks (in each axis) that the cov matrices consists of
    """
    assert Rb.shape == Rd.shape
    assert Rb.shape[0] == Rd.shape[1]
    assert Rb.ndim == 2
    #Rd += 1e-4*np.eye(Rd.shape[0])
    if reg_param > 0:
        Rd_reg = Rd + reg_param*np.eye(Rd.shape[0])
    else:
        Rd_reg = Rd
    eigvals, evec = splin.eigh(Rb, mat.ensure_pos_def_adhoc(Rd_reg, verbose=True))
    
    #ir = evec[:,-1].reshape(1, num_ls, -1)
    #norm = np.sqrt(np.sum(ir**2))
    #ir /= norm
    #ir *= 1e4
    return evec[:,-1] * np.sqrt(eigvals[-1])

def acc_time_all_zones(R, reg_param=0):
    """Time-domain acoustic contrast control for all zones

    Parameters
    ----------
    R : ndarray of shape (num_zones, num_zones, bf_len, bf_len)
        R[k,i,:,:] means spatial covariance associated with RIRs 
        of zone k, and audio signal of zone i
    
    Returns
    -------
    beamformer vector : ndarray of shape (num_zones, bf_len)
    """
    num_zones = R.shape[0]
    bf_len = R.shape[-1]
    assert R.shape == (num_zones, num_zones, bf_len, bf_len)

    Rd = np.zeros((bf_len, bf_len))
    w = np.zeros((num_zones, bf_len))

    for k in range(num_zones):
        Rd.fill(0)
        Rb = R[k,k,:,:]
        for i in range(num_zones):
            if i != k:
                Rd += R[i,k,:,:]
        w[k,:] = acc_time(Rb, Rd, reg_param)

    return w




# ================== LOUDSPEAKER SIGNAL TRANSFORM ================
def szc_transform_mwf(Rb, Rd, mu=1):
    """Calculates the linear transformation to apply to loudspeaker signals to obtain sound zones. 

    A special case of szc_transform_mwf_gevd, where rank is equal to the number of loudspeakers

    Parameters
    ----------
    Rb : ndarray of shape (num_freqs, num_ls, num_ls)
        spatial covariance associated with the bright zone
    Rd : ndarray of shape (num_freqs, num_ls, num_ls)
        spatial covariance associated with the dark zones
    mu : int
        non-negative, weights effort between error in dark and bright zones

    Returns
    -------
    W : ndarray of shape (num_freqs, num_ls, num_ls)
        The matrix representing the linear transformation of the loudspeaker signals. 
        Calculated as W = (Rb + mu*Rd)^{-1} Rb

    References
    ----------
    [brunnstromSound2023] J. Brunnström, T. van Waterschoot, and M. Moonen, “Sound zone control for arbitrary sound field reproduction methods,” in European Signal Processing Conference (EUSIPCO), Helsinki, Finland, Sep. 2023.
    """

    mat_to_invert = Rb + mu*Rd
    W = np.linalg.solve(mat_to_invert, Rb)
    return W

def szc_transform_mwf_gevd(Rb, Rd, rank, mu=1):
    """Calculates the linear transformation to apply to loudspeaker signals to obtain sound zones. 

    Parameters
    ----------
    Rb : ndarray of shape (num_freqs, num_ls, num_ls)
        spatial covariance associated with the bright zone
    Rd : ndarray of shape (num_freqs, num_ls, num_ls)
        spatial covariance associated with the dark zones
    mu : int
        non-negative, weights effort between error in dark and bright zones
    rank : int
        between 1 and num_ls. Applies low-rank approximation via GEVD. 
        Lower rank gives higher acoustic contrast, higher rank gives lower distortion

    Returns
    -------
    W : ndarray of shape (num_freqs, num_ls, num_ls)
        The matrix representing the linear transformation of the loudspeaker signals
    
    References
    ----------
    [brunnstromSound2023] J. Brunnström, T. van Waterschoot, and M. Moonen, “Sound zone control for arbitrary sound field reproduction methods,” in European Signal Processing Conference (EUSIPCO), Helsinki, Finland, Sep. 2023.
    """
    num_freqs = Rb.shape[0]
    W = np.zeros(Rb.shape, dtype=Rb.dtype)

    for f in range(num_freqs):
        eigvals, eigvec = splin.eigh(Rb[f,:,:], mu*Rd[f,:,:]) #largest eigenvalue is last
        eigvec = np.flip(eigvec, axis=-1)
        eigvec_invt = splin.inv(eigvec).T.conj()
        eigvals = np.flip(eigvals, axis=-1)
        eigvals = eigvals / (np.ones_like(eigvals) + eigvals)
        for r in range(rank):
            W[f,:,:] += eigvals[r] * eigvec[:,r:r+1] @ eigvec_invt[:,r:r+1].conj().T
    return W


# ================== SPATIAL AUDIO ===============================
def pressure_matching_sa(Hv, Hc):
    """Frequency domain pressure matching to generate loudspeaker signals from virtual source signals

    v is the virtual source sound pressure (num_freq, num_virtual_src, 1)
    
    Parameters
    ----------
    Hv : ndarray of shape (num_freq, num_mic, num_virtual_src)
        transfer functions from virtual sources to control points 
    Hc : ndarray of shape (num_freq, num_mic, num_ls)
        transfer functions from loudspeakers to control points 

    Returns
    -------
    beamformer w : ndarray of shape (num_freq, num_ls, num_virtual_source) 
        which should be applied to the virutal source sound pressure 
        as w @ v, where v : (num_freq, num_virtual_src, 1)

    Notes
    -----
    The beamformer is calculated as w = (H_c^H H_c)^{-1} H_c^H H_v. Definition can be found in (15) in [brunnstromSound2023]

    References
    ----------
    [brunnstromSound2023] J. Brunnström, T. van Waterschoot, and M. Moonen, “Sound zone control for arbitrary sound field reproduction methods,” in European Signal Processing Conference (EUSIPCO), Helsinki, Finland, Sep. 2023.
    """
    w = np.linalg.solve(np.moveaxis(Hc.conj(),-1,-2) @ Hc, np.moveaxis(Hc.conj(),-1,-2) @ Hv)
    return w

class PressureMatchingWOLA:
    """
    Sound field reproduction with pressure matching in the WOLA domain
    
    """
    def __init__(self, audio_src, Hv, Hc, block_size):
        self.ctrl_freq = self.update(Hv, Hc)
        self.audio_src = audio_src
        self.block_size = block_size

        self.wola = fc.WOLA(self.audio_src.num_channels, 1, 2*self.block_size, block_size)

        self.num_real_freq = Hv.shape[0]
        assert Hc.shape[0] == self.num_real_freq
        
    def get_samples(self):
        audio = self.audio_src.get_samples(self.block_size)
        self.wola.analysis(audio)
        ls_sig = self.ctrl_freq @ np.moveaxis(self.wola.spectrum, -1, 0)
        return ls_sig

    def update(self, Hv, Hc):
        self.ctrl_freq = pressure_matching_sa(Hv, Hc)







# ==================== SINR BASED SOUND ZONE CONTROL =============================
def solve_power_weighted_qos_uplink(R, noise_pow, sinr_targets, max_pow, audio_cov, tolerance=1e-12, max_iters=20, verbose=False):
    """ Calculates a control filter and power allocation that solves the signal-weighted uplink Quality of Service problem in the time-domain
    
    The control filter is SINR-optimal and takes the spectral characteristics of the audio signal into account. In order to use it 
    for sound zone control it should first be power-normalized with normalize_beamformer, and then scaled according to power_alloc_qos_downlink
    It is the proposed method of [brunnstromSignaltointerferenceplusnoise2023]. 
    
    Parameters
    ----------
    R : ndarray of shape (num_zones, num_zones, bf_len, bf_len)
        R[z,i,:,:] is R_{zi} from the paper below, can be seen in eq. (5)
        The first axis is the room impulse responses from loudspeakers to zone z
        The second axis is the desired audio signal for zone i 
    noise_power : ndarray of shape (num_zones,)
    sinr_targets : ndarray of shape (num_zones,)
    max_pow : float
    audio_cov : ndarray of shape (num_zones, bf_len, bf_len)
        defined in the paper below right above (7) or in equation (8)
    tolerance : float
        The relative mean square difference between two subsequent iterations at which 
        the algorithm is considered to have converged. 
    max_iters : int
        The number of iterations after which the algorithm terminates regardless of convergence.
        For infeasible max_pow and/or sinr_targets, the algorithm will never converge. 
        For feasible parameters, depending on the tolerance, the algorithm generally converges in 3-7 iterations

    Returns
    -------
    w : ndarray of shape (num_zones, bf_len)
        the SINR-optimal control filter
    q : ndarray of shape (num_zones,)
        the uplink power allocation
    
    References
    ----------
    [brunnstromSignaltointerferenceplusnoise2023] J. Brunnström, T. van Waterschoot, and M. Moonen, “Signal-to-interference-plus-noise ratio based optimization for sound zone control,” IEEE Open Journal of Signal Processing, vol. 4, pp. 257–266, 2023, doi: 10.1109/OJSP.2023.3246398.
    """
    num_zones = R.shape[0]

    q = np.zeros((num_zones))
    n = 0
    beamformers = []
    power_vectors = []

    is_feasible = False
    while True:
        if verbose:
            print(f"Iter {n}")
        w = _beamformer_minmax_weighted_uplink(q, R, audio_cov)
        w = normalize_beamformer(w)
        s = np.squeeze(np.array([w[k,:,None].T @ audio_cov[k,:,:] @ w[k,:,None] for k in range(num_zones)]))
        if is_feasible:
            q = _power_alloc_qos_uplink(w, R, s, sinr_targets)
        else:
            q, c = _power_alloc_minmax_uplink(w, R, s, sinr_targets, max_pow, verbose=verbose)
            if verbose:
                print(f"capacity: {c}")
            
            if _power_alloc_qos_feasibility_spectral_radius(_link_gain_uplink(w, R), sinr_targets) < 1:
                is_feasible = True
            
        beamformers.append(w)
        power_vectors.append(q)
        sinr_balance_diff = _sinr_balance_difference_uplink(q, w, R, noise_pow, sinr_targets)
        if verbose:
            print(f"SINR balance difference: {sinr_balance_diff}")
            print(f"Total power: {np.sum(q)}")
            print(f"Uplink feasibility spectral radius: {_power_alloc_qos_feasibility_spectral_radius(_link_gain_uplink(w, R), sinr_targets)}")
            print(f"Downlink feasibility spectral radius: {_power_alloc_qos_feasibility_spectral_radius(_link_gain_downlink(w, R), sinr_targets)}")
        if n >= 1:
            pow_diff = np.mean(np.abs(power_vectors[-1] - power_vectors[-2])**2)
            bf_diff = np.mean(np.abs(beamformers[-1] - beamformers[-2])**2)
            if verbose:
                print(f"Power mean square difference: {pow_diff}")
                print(f"Beamformer mean square difference: {bf_diff}")
            
            if (is_feasible and pow_diff < tolerance and bf_diff < tolerance) or n == max_iters:
                break
        n += 1
    return w, q

def sum_pow_of_mat(bf_mat):
    """
    bf_mat is the beamformer matrix obtained from semidefinite relaxation, 
        of shape (num_freqs, num_zones, num_sources, num_sources)

    return the sum of traces for each frequency, so an array of shape 
        (num_freqs,)
    """
    return np.real_if_close(np.sum(np.trace(bf_mat, axis1=-2, axis2=-1), axis=-1))

def sum_pow_of_vec(bf_vec):
    """

    Parameters
    ----------
    bf_vec : ndarray of shape (num_freqs, num_zones, num_sources)
        complex or real beamformer vector

    Returns
    -------
    The sum power for each frequency, so an array
        of shape (num_freqs,)
    
    """
    return np.sum(np.abs(bf_vec)**2, axis=(-2, -1))

def sum_weighted_pow(bf_vec, weighting_mat):
    """

    Parameters
    ----------
    bf_vec is complex or real beamformer vector of shape
        (num_zones, bf_len)
    weighting_mat is shape (num_zones, bf_len, bf_len)

    Returns
    -------
    Weighted sum power : float    
    """
    num_zones = bf_vec.shape[0]
    return np.sum([np.squeeze(bf_vec[k,:,None].T @ weighting_mat[k,:,:] @ bf_vec[k,:,None]) for k in range(num_zones)])

def _select_solution_eigenvalue(opt_mat, verbose=False):
    """
    opt_mat is of shape (..., beamformer_len, beamformer_len)

    returns vector of shape (..., beamformer_len)
    
    """
    assert opt_mat.shape[-2] == opt_mat.shape[-1]
    return mat.broadcast_func(opt_mat, _select_solution_eigenvalue_single_freq, out_shape=(opt_mat.shape[-1],), dtype=opt_mat.dtype, verbose=verbose)

def _select_solution_eigenvalue_single_freq(opt_mat, verbose=False):
    ev, evec = splin.eigh(opt_mat)
    if verbose:
        if ev[-1] / ev[-2] < 1e6:
                print(f"Ratio between 1st and 2nd eigval is {ev[-1] / ev[-2]}")
    return evec[:,-1] * np.sqrt(ev[-1])


def normalize_beamformer(w):
    """
    Normalizes the beamformer vector w_k
        so that each vector is length one, meaning ||w_k||_2 == 1
    """
    #w_norm = np.zeros_like(w)
    norm = splin.norm(w, axis=-1)
    w_norm = w / norm[...,None]
    return w_norm

def normalize_system(R, noise_pow):
    """
    Normalize spatial covariance matrix by noise powers to get
        unity noise power for all zones. Does not change the 
        downlink SINR, but does change uplink SINR. 

    Is desirable to use because strong duality is not guaranteed 
        when we have non-unity noise powers. 

    returns normalized_R, normalized_noise_pow
    """
    num_zones = noise_pow.shape[-1]
    assert noise_pow.ndim == 1
    assert len(noise_pow) == num_zones
    R_normalized = np.zeros_like(R)
    
    for k in range(num_zones):
        R_normalized[k,...] = R[k,...] / noise_pow[k]

    return R_normalized, np.ones_like(noise_pow)

def apply_power_vec(w, p):
    """
    w is (num_freq, num_zones, num_sources), or (num_zones, bf_len)
    """
    return np.sqrt(p[...,None]) * w

def extract_power_vec(w):
    """
    Opposite of apply_power_vec
    This function takes a non-unit-vector w and returns a unit vector w_norm
    along with the power p such that apply_power_vec(w_norm, p) = w

    w is (num_freq, num_zones, num_sources), or (num_zones, bf_len)
    """
    norm = splin.norm(w, axis=-1)
    w_norm = w / norm[...,None]
    return w_norm, norm**2

def _is_unit_vector(w):
    return np.allclose(np.linalg.norm(w, axis=-1), 1)

def _link_gain_downlink(w, R):
    return _link_gain(w, R)

def _link_gain_uplink(w, R):
    return _link_gain(w, R).T

def _link_gain(w, R):
    """
    R is (num_zones, bf_len, bf_len) or (num_zones, num_zones, bf_len, bf_len)
        if R.ndim == 4, the second index is the audio signal index. 

        returns the matrix G as defined in 
        'A general duality theory for uplink and downlink beamforming'
        which is defined as G_ik = w_k^T R_ik w_k
    """
    num_zones = w.shape[0]
    G = np.zeros((num_zones, num_zones))

    if R.ndim == 4:
        for k in range(num_zones):
            for i in range(num_zones):
                G[i,k] = np.real_if_close(np.squeeze(w[k,:,None].T.conj() @ R[i,k,:,:] @ w[k,:,None]))
    if R.ndim == 3:
        for k in range(num_zones):
            for i in range(num_zones):
                G[i,k] = np.real_if_close(np.squeeze(w[k,:,None].T.conj() @ R[i,:,:] @ w[k,:,None]))
    return G


def _sinr_downlink(p, w, R, noise_pow):
    assert _is_unit_vector(w)
    return _sinr(p, _link_gain_downlink(w, R), noise_pow)

def _sinr_uplink(p, w, R, noise_pow):
    assert _is_unit_vector(w)
    return _sinr(p, _link_gain_uplink(w, R), noise_pow)

def _sinr_margin_downlink(p, w, R, noise_pow, sinr_targets):
    return _sinr_downlink(p, w, R, noise_pow) - sinr_targets

def _sinr_margin_uplink(p, w, R, noise_pow, sinr_targets):
    return _sinr_uplink(p, w, R, noise_pow) - sinr_targets

def _sinr(p, gain_mat, noise_pow):
    """
    if gain_mat is obtained from the function link_gain
        then gain_mat = link_gain() matches the downlink SINR
        and gain_mat = link_gain().T matches the uplink SINR

    gain_mat is of shape (num_zones, num_zones)
    noise_pow is of shape (num_zones)

    returns a SINR for each zone, an array of shape (num_zones,)
    """
    assert p.ndim == 1
    assert gain_mat.ndim == 2
    assert gain_mat.shape[0] == gain_mat.shape[1]
    num_zones = gain_mat.shape[0]
    assert len(p) == num_zones
    assert len(noise_pow) == num_zones

    gain_mat_scaled = gain_mat @ np.diag(p)
    interference = _sum_interference(_interference_matrix(gain_mat_scaled))

    sinr_val = np.zeros((num_zones))
    for k in range(num_zones):
        sinr_val[k] = gain_mat_scaled[k,k] / (interference[k] + noise_pow[k])
    return sinr_val

def _sum_interference(interference_mat):
    return np.sum(interference_mat, axis=-1)
        
    
def power_alloc_qos_downlink(w, R, noise_pow, sinr_targets):
    assert _is_unit_vector(w)
    return _power_alloc_qos(_link_gain_downlink(w, R), noise_pow, sinr_targets)

def _power_alloc_qos_uplink(w, R, noise_pow, sinr_targets):
    assert _is_unit_vector(w)
    return _power_alloc_qos(_link_gain_uplink(w, R), noise_pow, sinr_targets)

def _power_alloc_qos(gain_mat, noise_pow, sinr_targets):
    """
    Minimizes the power when the SINR equals the sinr_targets

    Closed form derivations and proofs given in 
        'A general duality theory for uplink and downlink beamforming'
    
    """
    assert _power_alloc_qos_is_feasible(gain_mat, sinr_targets)
    if_mat = _interference_matrix(gain_mat)
    sig_mat = _signal_diag_matrix(gain_mat, sinr_targets)

    system_mat = np.eye(gain_mat.shape[-1]) - sig_mat @ if_mat
    answer_mat = sig_mat @ noise_pow[:,None]
    p = splin.solve(system_mat, answer_mat)

    assert np.all(p > 0)
    return p[:,0]

def _power_alloc_qos_is_feasible(gain_mat, sinr_targets):
    """
    Derivations in 'A General Duality Theory for Uplink and Downlink Beamforming'
    """
    return _power_alloc_qos_feasibility_spectral_radius(gain_mat, sinr_targets) < 1

def _power_alloc_qos_feasibility_spectral_radius(gain_mat, sinr_targets):
    """
    Derivations in 'A General Duality Theory for Uplink and Downlink Beamforming'
    """
    if_mat = _interference_matrix(gain_mat)
    sig_mat = _signal_diag_matrix(gain_mat, sinr_targets)
    ev = splin.eigvals(sig_mat @ if_mat)
    spectral_radius = np.max(np.abs(ev))
    return spectral_radius

def _power_alloc_minmax_downlink(w, R, noise_pow, sinr_targets, max_pow, verbose=False):
    assert _is_unit_vector(w)
    return _power_alloc_minmax(_link_gain_downlink(w, R), noise_pow, sinr_targets, max_pow, verbose=verbose)

def _power_alloc_minmax_uplink(w, R, noise_pow, sinr_targets, max_pow, verbose=False):
    assert _is_unit_vector(w)
    return _power_alloc_minmax(_link_gain_uplink(w, R), noise_pow, sinr_targets, max_pow, verbose=verbose)

def _power_alloc_minmax(gain_mat, noise_pow, sinr_targets, max_pow, verbose=False):
    """
    Tranpose the gain_mat to shift between the downlink and uplink solution
    
    """
    if_mat = _interference_matrix(gain_mat)
    D = _signal_diag_matrix(gain_mat, sinr_targets)
    coupling_mat = _extended_coupling_matrix_downlink(D, if_mat, noise_pow, max_pow)
    eigvals, eigvec = splin.eig(coupling_mat)
    #print(eigvals)

    max_ev_idx = np.argmax(np.real(eigvals))
    pow_vec = np.real_if_close(eigvec[:-1, max_ev_idx] / eigvec[-1, max_ev_idx])
    assert np.all(pow_vec > 0) #the dominant eigenvector should be positive
    if verbose:
        if not np.all([eigvals[i] <= 1e-10 for i in range(len(eigvals)) if i != max_ev_idx]): #only one eigenvalue should be positive (not sure if this should be true actually)
            print(f"Warning: Not only one eigenvalue is positive: {eigvals}")
    capacity = np.real_if_close(1 / eigvals[max_ev_idx])
    return pow_vec, capacity

def _interference_matrix(gain_mat):
    """
    The inference matrix is the link gain but with diagonal elements
        set to zero. 
        
    According to the definition written at the function 
        link_gain, to get the sum interference for a zone k, you take 
        np.sum(if_mat, axis=0)[k] (to be verified)
    """
    if_mat = np.zeros_like(gain_mat)
    if_mat[...] = gain_mat
    np.fill_diagonal(if_mat, 0)
    return if_mat

def _signal_diag_matrix(gain_mat, sinr_targets):
    return np.diag(sinr_targets / np.diag(gain_mat))

def _extended_coupling_matrix_downlink(signal_mat, if_mat, noise_pow, max_pow):
    """
    See documentation for power_assignment_minmax_downlink
    Returns equation (12) from Schubert & Boche
    """
    num_zones = signal_mat.shape[0]
    ext_dim = num_zones + 1

    noise_pow = noise_pow[:,None]           #make into column vector
    max_pow_vec = (1 / max_pow) * np.ones((1, num_zones))

    coupling_mat = np.zeros((ext_dim, ext_dim))
    coupling_mat[:-1, :-1] = signal_mat @ if_mat    #top left block
    coupling_mat[:-1, -1:] = signal_mat @ noise_pow                         #top right block
    coupling_mat[-1:, :-1] = max_pow_vec @ signal_mat @ if_mat                      #bottom left block
    coupling_mat[-1:, -1:] = max_pow_vec @ signal_mat @ noise_pow                          #bottom right block
    return coupling_mat

def _sinr_balance_difference_downlink(p, w, R, noise_pow, sinr_targets):
    return _sinr_balance_difference(p, _link_gain_downlink(w, R), noise_pow, sinr_targets)

def _sinr_balance_difference_uplink(p, w, R, noise_pow, sinr_targets):
    return _sinr_balance_difference(p, _link_gain_uplink(w, R), noise_pow, sinr_targets)

def _sinr_balance_difference(p, gain_mat, noise_pow, sinr_targets):
    sinr_val = _sinr(p, gain_mat, noise_pow)
    sinr_ratio  = sinr_targets / sinr_val
    return np.max(sinr_ratio) - np.min(sinr_ratio)

def _beamformer_minmax_weighted_uplink(q, R, audio_cov):
    num_zones = R.shape[0]
    bf_len = R.shape[-1]

    w = np.zeros((num_zones, bf_len), dtype=R.dtype)
    Q = np.zeros((bf_len, bf_len), dtype=R.dtype)

    for k in range(num_zones):
        Q.fill(0)
        Q += audio_cov[k,:,:]
        for i in range(num_zones):
            if k != i:
                if R.ndim == 3:
                    Q += q[i] * R[i,:,:]
                elif R.ndim == 4:
                    Q += q[i] * R[i,k,:,:]
        if R.ndim == 3:
            eigval, eigvec = splin.eigh(R[k,:,:], Q)
        elif R.ndim == 4:
            eigval, eigvec = splin.eigh(R[k,k,:,:], Q)
        w[k,:] = eigvec[:,-1]
    return w


"""Collection of algorithms for sound field estimation

* Kernel interpolation [uenoKernel2018]
* Infinite dimensional spherical harmonic analysis for moving microphones [brunnstromBayesianSubmitted]
* Spatial spectrum estimation for moving microphones [katzbergSpherical2021]

References
----------
[uenoKernel2018] N. Ueno, S. Koyama, and H. Saruwatari, “Kernel ridge regression with constraint of Helmholtz equation for sound field interpolation,” in 2018 16th International Workshop on Acoustic Signal Enhancement (IWAENC), Tokyo, Japan: IEEE, Sep. 2018, pp. 436–440. doi: 10.1109/IWAENC.2018.8521334. `[link] <https://doi.org/10.1109/IWAENC.2018.8521334>`__ \n
[brunnstromBayesianSubmitted] J. Brunnström, M. B. Møller, and M. Moonen, “Bayesian sound field estimation using moving microphones,” IEEE Open Journal of Signal Processing, submitted. \n
[katzbergSpherical2021] F. Katzberg, M. Maass, and A. Mertins, “Spherical harmonic representation for dynamic sound-field measurements,” in ICASSP 2021 - 2021 IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP), Jun. 2021, pp. 426–430. doi: 10.1109/ICASSP39728.2021.9413708. `[link] <https://doi.org/10.1109/ICASSP39728.2021.9413708>`__ \n
"""
import numpy as np
import scipy.spatial.distance as spdist

import aspcore.pseq as pseq
import aspcore.fouriertransform as ft

import aspcol.kernelinterpolation as ki
import aspcol.sphericalharmonics as sph




#============= FREQUENCY DOMAIN METHODS - STATIONARY MICROPHONES =============
def est_ki_diffuse_freq(p_freq, pos, pos_eval, k, reg_param):
    """Estimates the RIR in the frequency domain using kernel interpolation
    
    Uses the frequency domain sound pressure as input

    Parameters
    ----------
    p_freq : ndarray of shape (num_real_freqs, num_mics)
        sound pressure in frequency domain at num_mic microphone positions
    pos : ndarray of shape (num_mic, 3)
        positions of the microphones
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    k : ndarray of shape (num_freq)
        wavenumbers
    reg_param : float
        regularization parameter for kernel interpolation

    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        estimated RIR per frequency at the evaluation points

    References
    ----------
    [uenoKernel2018]
    """
    est_filt = ki.get_krr_parameters(ki.kernel_helmholtz_3d, reg_param, pos_eval, pos, k)
    p_ki = est_filt @ p_freq[:,:,None]
    return np.squeeze(p_ki, axis=-1)


def nearest_neighbour_freq(p_freq, pos, pos_eval):
    """
    Estimates the sound field at the evaluation points by simply selecting the value
    associated with the nearest microphone position. 

    Parameters
    ----------
    p_freq : ndarray of shape (num_real_freqs, num_mics)
        sound pressure in frequency domain at num_mic microphone positions
    pos : ndarray of shape (num_mic, 3)
        positions of the microphones
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points

    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        estimated RIR per frequency at the evaluation points
    """
    dist = spdist.cdist(pos, pos_eval)
    min_idx = np.argmin(dist, axis=0)

    num_real_freqs = p_freq.shape[0]
    est_sound_pressure = np.zeros((num_real_freqs, pos_eval.shape[0]), dtype=complex)
    for pos_idx in range(pos_eval.shape[0]):
        est_sound_pressure[:,pos_idx] = p_freq[:,min_idx[pos_idx]]
    return est_sound_pressure


def inf_dim_shd_analysis(p_freq, pos, pos_eval, wave_num, dir_coeffs, reg_param):
    """Estimates the sound field using directional microphones 

    Since we reconstruct the sound pressure immediately, without explicitly computing
    the spherical harmonic coefficients, no truncation or expansion center must be set.
    
    Parameters
    ----------
    p_freq : ndarray of shape (num_freqs, num_mics)
        sound pressure in frequency domain at num_mic microphone positions
    pos : ndarray of shape (num_mic, 3)
        positions of the microphones
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    wave_num : ndarray of shape (num_freq)
        wavenumbers
    dir_coeffs : ndarray of shape (num_freq, num_mic, num_coeffs)
        harmonic coefficients of the directionality of the microphones
    reg_param : float
        regularization parameter. Must be non-negative
    
    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        estimated RIR per frequency at the evaluation points
    """
    num_mic = pos.shape[0]

    psi = sph.translated_inner_product(pos, pos, dir_coeffs, dir_coeffs, wave_num)
    psi_plus_noise_cov = psi + np.eye(num_mic) * reg_param
    regression_vec = np.linalg.solve(psi_plus_noise_cov, p_freq)[...,None]

    omni_dir = sph.directivity_omni()
    estimator_matrix = sph.translated_inner_product(pos_eval, pos, omni_dir, dir_coeffs, wave_num)
    est_sound_pressure = np.squeeze(estimator_matrix @ regression_vec, axis=-1)
    return est_sound_pressure
# ============= TIME DOMAIN METHODS - STATIONARY MICROPHONES =============


def pseq_nearest_neighbour(p, seq, pos, pos_eval):
    """
    Estimates the sound field at the evaluation points by simply selecting the value
    associated with the nearest microphone position. 
    Assumes that the sequence is a perfect periodic sequence.

    Parameters
    ----------
    p : ndarray of shape (num_mic, seq_len)
        sound pressure in time domain at num_mic microphone positions
    seq : ndarray of shape (seq_len) or (1, seq_len)
        training signal used for the measurements
    pos : ndarray of shape (num_mic, 3)
        positions of the microphones
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points

    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        estimated RIR per frequency at the evaluation points
    """
    rir = pseq.decorrelate(p, seq)
    rir_freq = ft.rfft(rir)

    dist = spdist.cdist(pos, pos_eval)
    min_idx = np.argmin(dist, axis=0)

    num_real_freqs = rir_freq.shape[0]
    est_sound_pressure = np.zeros((num_real_freqs, pos_eval.shape[0]), dtype=complex)
    for pos_idx in range(pos_eval.shape[0]):
        est_sound_pressure[:,pos_idx] = rir_freq[:,min_idx[pos_idx]]
    return est_sound_pressure


def est_ki_diffuse(p, seq, pos, pos_eval, samplerate, c, reg_param):
    """
    Estimates the RIR in the frequency domain using kernel interpolation
    Assumes seq is a perfect periodic sequence

    Parameters
    ----------
    p : ndarray of shape (M, seq_len)
        sound pressure in time domain at M microphone positions
    seq : ndarray of shape (seq_len)
        the training signal used for the measurements
    pos : ndarray of shape (num_mic, 3)
        positions of the microphones
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    samplerate : int
    c : float
        speed of sound
    reg_param : float
        regularization parameter for kernel interpolation

    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        estimated RIR per frequency at the evaluation points

    References
    ----------
    [uenoKernel2018]
    """
    rir = pseq.decorrelate(p, seq)

    fft_len = rir.shape[-1]
    rir_freq = ft.rfft(rir)
    k = ft.get_real_wavenum(fft_len, samplerate, c)

    return est_ki_diffuse_freq(rir_freq, pos, pos_eval, k, reg_param)













"""Module for working with plane waves and plane wave models.

The plane wave is defined as $e^{-ik(r-r_c)^T d}$ where r is the position, d is the direction of the plane wave.
and r_c is the expansion center (the point around which the directions are calculated). 

Using the time-harmonic convention of $exp(-iwt)$, the plane wave is defined as $exp(ikr^T d)$ where d is the plane 
wave propagation direction [martinMultiple2006]. Therefore the direction provided is the direction from which the plane wave is incoming.

References
----------
[martinScattering2006] P. A. Martin, Multiple scattering: Interaction of time-harmonic waves with N obstacles, vol. 107. in Encyclopedia of mathematics and its applications, vol. 107. Cambridge, UK: Cambridge University Press, 2006.

"""

import numpy as np
import scipy.special as special
import aspcore.montecarlo as mc

import aspcol.utilities as utils
import aspcol.sphericalharmonics as sph



def plane_wave(pos, direction, wave_num, exp_center=None):
    """The complex response of a plane wave for a specific frequency for a set of positions.

    Implements exp(-ik(r-r_c)^T d) where r is the position, d is the direction of the plane wave.
    and r_c is the expansion center (the point around which the directions are calculated). 

    Using the time-harmonic convention of exp(-iwt), the plane wave is defined as exp(ikr^T d) where d is the plane 
    wave propagation direction [martinMultiple2006]. Therefore the direction provided is the direction from which the plane wave is incoming.

    pos : ndarray of shape (num_positions, 3)
        The positions where the plane wave is evaluated
    direction : ndarray of shape (num_direction, 3)
        The directionfrom which the plane wave is incoming. Must be a unit vector
    wave_num : float or ndarray of shape (num_real_freqs,)
        The wave number of the plane wave. Defined as 2*pi*f/c where f is the frequency 
        and c is the speed of sound.

    Returns
    -------
    plane_wave : ndarray of shape (num_positions, num_direction) or (num_real_freqs, num_positions, num_direction)
        The complex response of the plane wave at the positions. 

    References
    ----------
    [martinMultiple2006] P. A. Martin, Multiple scattering: Interaction of time-harmonic waves with N obstacles, vol. 107. in Encyclopedia of mathematics and its applications, vol. 107. Cambridge, UK: Cambridge University Press, 2006.
    """
    if direction.ndim == 1:
        direction = direction[None,:]
    assert direction.ndim == 2
    assert direction.shape[-1] == 3
    assert np.allclose(np.linalg.norm(direction, axis=-1), 1)
    assert pos.ndim == 2
    assert pos.shape[1] == 3

    if exp_center is not None:
        if exp_center.ndim == 1:
            exp_center = exp_center[None,:]
        assert exp_center.shape == (1,3)
        pos = pos - exp_center

    if np.isscalar(wave_num):
        wave_num = np.array([wave_num])
    assert wave_num.ndim == 1
    num_freqs = wave_num.shape[0]

    pw_values = np.exp(-1j * wave_num[:,None,None] * np.sum(pos[:,None,:] * direction[None,:,:], axis=-1)[None,...])

    if num_freqs == 1:
        pw_values = np.squeeze(pw_values, axis=0)
    return pw_values



def find_matching_plane_wave(plane_wave_data, freq, pw_dir, pos_mic, c):
    """Matches a plane wave to the given data

    The function returns directly the sound pressure of the matched plane wave, but also
    enough information for the plane wave to be recreated as gain_adjustment * pw.plane_wave(pos_mic, pw_dir, wave_num, exp_center)

    Parameters
    ----------
    plane_wave_data : ndarray of shape (num_mic, 1)
        data from a plane wave or plane-wave like sound field that should be matched by the analytic plane wave
    freq : float
        frequency of the plane wave
    pw_dir : ndarray of shape (1, 3)
        Direction from which the plane wave is incoming. Must be a unit vector
    pos_mic : ndarray of shape (num_mic, 3)
        positions of the microphones (the points where the plane wave and the plane_wave_data are measured)
    c : float
        speed of sound

    Returns
    -------
    p_mic_analytic : ndarray of shape (num_mic, 1)
        the sound pressure at pos_mic of the analytic plane wave that matches the plane_wave_data. 
        If plane_wave_data is a plane wave or very close, then p_mic_analytic should be 
        essentially identical to plane_wave_data
    exp_center : ndarray of shape (1, 3)
        the expansion center of the plane wave
    gain_adjustment : float
        A gain adjustment that should be applied to the analytic plane wave to match the plane_wave_data
    
        
    Notes
    -----
    The gain adjustment is found as the mean amplitude of the plane_wave_data. 

    The phase adjustment is found as the solution to $\\lVert e^{i \\gamma} \\bm{p} - \\bm{s} \\rVert_2^2$, where $\\bm{p}$ is 
    the vector of analytic pressure values and $\\bm{s}$ is the vector of plane_wave_data. The solution is found analytically
    by taking the derivative and setting to zero. 

    The expansion center is defined by $exp(-ik(r-r_c)^T d) = exp(-i gamma) exp(ikr^T d)$, where r_c is the unknown expansion center
    and the phase adjustment gamma is found by the algorithm. Since that relationship is not necessarily unique, the 
    expansion center is found as min \\lVert r_c \\rVert_2^2 subject to r_c^T d = gamma / wave_num, which is a 
    least squares problem with linear constrained, solved analytically using the Lagrange multiplier method.
    """
    freq = np.ones(1) * freq
    wave_num = 2 * np.pi * freq / c
    p_mic_analytic = np.squeeze(plane_wave(pos_mic, pw_dir, wave_num), axis=-1)
    #p_mic_analytic = np.conj(p_mic_analytic) # to change time convention

    sim_amplitude = np.mean(np.abs(plane_wave_data))
    analytic_amplitude = np.mean(np.abs(p_mic_analytic))
    gain_adjustment = sim_amplitude / analytic_amplitude
    p_mic_analytic = p_mic_analytic * gain_adjustment

    phase_adjustment_angle = -(1j / 2) * (np.log(np.sum(np.conj(p_mic_analytic) * plane_wave_data)) - np.log(np.sum(p_mic_analytic * np.conj(plane_wave_data))) )
    phase_adjustment = np.exp(1j * phase_adjustment_angle)
    p_mic_analytic = p_mic_analytic * phase_adjustment

    exp_center = np.real_if_close(pw_dir * phase_adjustment_angle / wave_num)
    return p_mic_analytic, exp_center, gain_adjustment

def shd_coeffs_for_planewave(pw_direction, max_order):
    """Spherical harmonic coefficients for a plane wave exp(-ikr^T d)
    where r is the position and d is the direction of the plane wave.

    The expansion center is assumed to be at the origin.

    Parameters
    ----------
    pw_direction : ndarray of shape (num_pw, 3)
        The direction of the plane wave. Must be a unit vector
    max_order : int
        The maximum order of the spherical harmonics expansion    
    
    Returns
    -------
    shd_coeffs : ndarray of shape (num_pw, num_coeffs)
        The spherical harmonic coefficients for the plane wave
    """
    assert pw_direction.ndim == 2
    assert pw_direction.shape[1] == 3
    assert np.allclose(np.linalg.norm(pw_direction, axis=-1), 1)

    rad, angles = utils.cart2spherical(pw_direction)
    orders, degrees = sph.shd_num_degrees_vector(max_order)
    const_factor = (-1j)**orders * np.sqrt(4 * np.pi)
    sph_harm = special.sph_harm(degrees[None,:], orders[None,:], angles[...,0,None], angles[...,1,None])
    shd_coeffs = const_factor[None,:] * np.conj(sph_harm)
    return shd_coeffs

def _shd_coeffs_for_planewave_positive(pw_direction, max_order):
    """Harmonic coefficients for a plane wave exp(ikr^T d)
    where r is the position and d is the direction of the plane wave.

    The expansion center is assumed to be at the origin.

    Parameters
    ----------
    pw_direction : ndarray of shape (num_pw, 3)
        The direction of the plane wave. Must be a unit vector
    max_order : int
        The maximum order of the spherical harmonics expansion    
    
    Returns
    -------
    shd_coeffs : ndarray of shape (num_pw, num_coeffs)
        The spherical harmonic coefficients for the plane wave
    """
    assert pw_direction.ndim == 2
    assert pw_direction.shape[1] == 3
    assert np.allclose(np.linalg.norm(pw_direction, axis=-1), 1)

    rad, angles = utils.cart2spherical(pw_direction)
    orders, degrees = sph.shd_num_degrees_vector(max_order)

    # Calculate harmonic coefficients
    const_factor = 1j**orders * np.sqrt(4 * np.pi)
    sph_harm = special.sph_harm(degrees[None,:], orders[None,:], angles[...,0,None], angles[...,1,None])
    shd_coeffs = const_factor[None,:] * np.conj(sph_harm)
    return shd_coeffs


def plane_wave_integral(dir_func, pos, exp_center, wave_num, rng, num_samples):
    """Computes the integral of a function multiplied with a plane wave over a sphere.

    Defined according to (6) in Brunnström et al 2024.
    int_{S^2} f(d) exp(-ik(r-r_c)^T d) ds(d)
    where S^2 is the unit sphere, f(d) is the function, r is the position, r_c is the expansion center,	
    d is the incoming direction of the plane wave, k is the wave number, and ds(d) is the surface element 
    of the sphere.

    Parameters
    ----------
    dir_func : function
        A function that takes direction unit vectors, ndarray of shape (num_points, 3)
        and returns a complex value response of shape (num_points)
    pos : ndarray of shape (num_pos, 3)
        The position where 
    exp_center : ndarray of shape (1,3)
        The center of the expansion
    wave_num : float
        The wave number of the plane wave
    rng : numpy.random.Generator
        The random number generator to use

    Returns 
    -------
    est : ndarray of shape (num_pos,)
        The estimated value of the integral evaluated at all the supplied positions

    Notes
    -----
    Same definition is (9) in Ribeiro 2023, but with a sign difference in the complex exponential,
    and without an expansion center. 

    Performs the integration using Monte Carlo integration. For computational cost sensitive applications,
    there are more efficient methods available, such as using a Lebedev quadrature.
    """
    dir_vecs = mc.uniform_random_on_sphere(num_samples, rng)

    func_values = dir_func(dir_vecs)
    planewave_values = plane_wave(pos - exp_center, dir_vecs, wave_num)
    mean_integrand = np.mean(func_values[None,:] * planewave_values, axis=-1)

    sphere_area = 4 * np.pi # must multiply by area of integration domain
    est = sphere_area * mean_integrand
    return est








# ==============================================================================
# MEASUREMENTS WITH PLANE WAVE MODELS

def omni_directivity_function():
    """

    """
    def dir_func(dir_vec):
        """
        dir_vec : ndarray of shape (num_directions, 3)
        these are the directions for which we test what the
        microphone response is. 
        """
        assert dir_vec.ndim == 2
        assert dir_vec.shape[1] == 3
        return np.ones((dir_vec.shape[0],), dtype=complex)
    return dir_func

def linear_directivity_function(A, d_mic):
    """
    
    Omni directivity is obtained by setting A = 0
    Cardoid directivity is obtained by setting A = 1/2
    Figure-8 directivity is obtained by setting A = 1

    d_mic : ndarray of shape (3,) or (1, 3)
        direction microphone is pointing, the peak directivity of e.g. a cardioid mic. 
    """
    if d_mic.ndim == 1:
        d_mic = d_mic[None,:]
    assert d_mic.ndim == 2
    assert d_mic.shape[1] == 3
    assert 0 <= A <= 1

    def dir_func(dir_vec):
        """
        dir_vec : ndarray of shape (num_directions, 3)
        these are the directions for which we test what the
        microphone response is. 
        """
        assert dir_vec.ndim == 2
        assert dir_vec.shape[1] == 3
        #dir_vec = np.expand_dims(dir_vec, axis=1)
        return (1-A) + A * np.sum(dir_vec * d_mic, axis=-1, dtype=complex)
    return dir_func



def apply_measurement(pw_coeffs, dir_func, rng, num_samples):
    """Returns the signal that a microphone with directionality given by dir_func would record. 
    The microphone must be located at the expansion center for the plane wave coefficients.

    Defined as (7) in Brunnström et al 2024.

    Parameters
    ----------
    pw_coeffs : function
        the coefficients of the plane wave expansion defining the soundfield. This is a function that
        takes a direction unit vector and returns a complex value response.
    dir_func : function
        A function that takes direction unit vectors, ndarray of shape (num_points, 3)
        and returns a complex value response of shape (num_points)
    rng : numpy.random.Generator
        The random number generator to use
    num_samples : int
        The number of samples to use for the monte carlo integration

    Returns
    -------
    
    """
    dir_vecs = mc.uniform_random_on_sphere(num_samples, rng)
    
    pw_vals = pw_coeffs(dir_vecs)
    dir_vals = dir_func(dir_vecs)
    mean_integrand = np.mean(pw_vals * dir_vals, axis=-1)
    
    sphere_area = 4 * np.pi
    est = sphere_area * mean_integrand
    return est


"""Module for recording sound fields using a moving microphone

The estimation of a sound field from moving microphones is very computationally costly. Especially for directional microphones, the computational cost of the estimation can be prohibitive. Therefore a lot of the code in this module is implemented in jax, such that the resulting functions can be compiled, leading to considerable improvements in running time. 

Due to the need of re-implementing functions such as the spherical Bessel function in jax, the current compilable implementations in this module are somewhat restricted. There is jax implementations of e.g. the translation operator that can also be found in the module sphericalharmonics.py, but in this module it assumes order 0 and 1 harmonic coefficients only. 

The sound field estimation function inf_dimensional_shd_dynamic cannot deal with directionalities above order 1.

References
----------
[brunnstromBayesianSubmitted] J. Brunnström, M. B. Møller, and M. Moonen, “Bayesian sound field estimation using moving microphones,” IEEE Open Journal of Signal Processing, submitted. \n
"""

import numpy as np
import scipy.linalg as splin
import scipy.special as spspec
import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
#jax.config.update("jax_disable_jit", True)
#jax.config.update("jax_debug_nans", True)

import aspcore.fouriertransform as ft

import aspcol.sphericalharmonics as shd
import aspcol.utilities as util
import aspcol.kernelinterpolation as ki





#============= MOVING MICROPHONE ESTIMATION - MOVED FROM SOUNDFIELDESIMTION.PY =============


def est_inf_dimensional_shd_omni(p, pos, pos_eval, sequence, samplerate, c, reg_param, verbose=False):
    """Estimates the RIR at evaluation positions using data from a moving microphone
    using Bayesian inference of an infinite sequence of spherical harmonics

    Assumptions:
    The microphones are omnidirectional
    The noise covariance is a scaled identity matrix
    The data is measured over an integer number of periods of the sequence
    N = seq_len * M, where M is the number of periods that was measured
    The length of sequence is the length of the estimated RIR

    Parameters
    ----------
    p : ndarray of shape (N)
        sound pressure for each sample of the moving microphone
    pos : ndarray of shape (N, 3)
        position of the trajectory for each sample
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    sequence : ndarray of shape (seq_len) or (1, seq_len)
        the training signal used for the measurements
    samplerate : int
    c : float
        speed of sound
    reg_param : float
        regularization parameter
    verbose : bool, optional
        if True, returns diagnostics, by default False

    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        estimated RIR per frequency at the evaluation points

    References
    ----------
    [brunnstromBayesianSubmitted]
    """
    # ======= Argument parsing =======
    if p.ndim >= 2:
        p = np.squeeze(p)

    if sequence.ndim == 2:
        sequence = np.squeeze(sequence, axis=0)
    assert sequence.ndim == 1

    assert pos.ndim == 2

    N = p.shape[0]
    seq_len = sequence.shape[0]
    num_periods = N // seq_len
    assert N % seq_len == 0

    k = ft.get_wavenum(seq_len, samplerate, c)
    num_real_freqs = len(ft.get_real_freqs(seq_len, samplerate))

    # ======= Estimation of spherical harmonic coefficients =======
    Phi = _sequence_stft_multiperiod(sequence, num_periods)

    #division by pi is a correction for the sinc function used later
    dist_mat = np.sqrt(np.sum((np.expand_dims(pos,1) - np.expand_dims(pos,0))**2, axis=-1))  / np.pi 
    
    psi = np.zeros((N, N), dtype = float)

    # no conjugation required for zeroth frequency and the Nyquist frequency, 
    # since they will be real already for a real input sequence
    psi += np.sinc(dist_mat * k[0]) * np.real_if_close(Phi[0,:,None] * Phi[0,None,:])
    assert seq_len % 2 == 0 #following line is only correct if B is even
    psi += np.sinc(dist_mat * k[seq_len//2]) * np.real_if_close(Phi[seq_len//2,:,None] * Phi[seq_len//2,None,:])

    for f in range(1, num_real_freqs-1):
        phi_rank1_matrix = Phi[f,:,None] * Phi[f,None,:].conj()
        psi += 2*np.real(np.sinc(dist_mat * k[f]) * phi_rank1_matrix)

    noise_cov = reg_param * np.eye(N)
    regressor = splin.solve(psi + noise_cov, p, assume_a = "pos")

    regressor = Phi.conj() * regressor[None,:]
    regressor = regressor[:num_real_freqs,:] # should be replaced by fixing the function calculating Phi instead

    # ======= Reconstruction of RIR =======
    est_sound_pressure = np.zeros((num_real_freqs, pos_eval.shape[0]), dtype=complex)
    for f in range(num_real_freqs):
        kernel_val = ki.kernel_helmholtz_3d(pos_eval, pos, k[f:f+1]).astype(complex)[0,:,:]
        est_sound_pressure[f, :] = np.sum(kernel_val * regressor[f,None,:], axis=-1)

    if verbose:
        #diagnostics = {}
        #diagnostics["regularization parameter"] = reg_param
        #diagnostics["condition number"] = np.linalg.cond(psi).tolist()
        #diagnostics["smallest eigenvalue"] = splin.eigh(psi, subset_by_index=(0,0), eigvals_only=True).tolist()
        #diagnostics["largest eigenvalue"] = splin.eigh(psi, subset_by_index=(N-1, N-1), eigvals_only=True).tolist()
        return est_sound_pressure, regressor, psi#, diagnostics
    else:
        return est_sound_pressure



def est_spatial_spectrum_dynamic(p, pos, pos_eval, sequence, samplerate, c, reg_param, r_max=None, verbose=False):
    """
    Estimates the RIR at evaluation positions using data from a moving microphone

    Implements the method from Katzberg et al. "Spherical harmonic 
    representation for dynamic sound-field measurements"
    
    Assumptions:
    The spherical harmonics are expanded around the origin (0,0,0)
    The sequence is periodic, and the pressure is measured for an integer number of periods
    The length of sequence is the length of the estimated RIR

    Parameters
    ----------
    p : ndarray of shape (N)
        sound pressure for each sample of the moving microphone
    pos : ndarray of shape (N, 3)
        position of the trajectory for each sample
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    sequence : ndarray of shape (seq_len) or (1, seq_len)
        the training signal used for the measurements
    samplerate : int
    c : float
        speed of sound
    r_max : float, optional
        radius of the sphere onto which the spatial spectrum is computed. If not provided, it is 
        set to the maximum distance from the origin to any of the microphone positions. 
    verbose : bool, optional
        if True, returns diagnostics, by default False

    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        estimated RIR per frequency at the evaluation points
    """
    # ======= Argument checking =======
    if p.ndim >= 2:
        p = np.squeeze(p)

    if sequence.ndim == 2:
        sequence = np.squeeze(sequence, axis=0)
    assert sequence.ndim == 1

    N = p.shape[0]
    seq_len = sequence.shape[0]
    num_periods = N // seq_len
    assert N % seq_len == 0
    num_eval = pos_eval.shape[0]

    k = ft.get_wavenum(seq_len, samplerate, c)
    num_freqs = len(k)
    num_real_freqs = len(ft.get_real_freqs(seq_len, samplerate))

    r, angles = util.cart2spherical(pos)

    # ======= Estimation of spatial spectrum coefficients =======
    phi = _sequence_stft_multiperiod(sequence, num_periods)

    if r_max is None:
        r_max = np.max(r)
    max_orders = shd.shd_min_order(k[:num_real_freqs], r_max)
    max_orders = np.concatenate((max_orders, np.flip(max_orders[1:])))
    
    Sigma = []

    freq_idx_list = np.arange(num_real_freqs)
    for f in freq_idx_list:#range(num_real_freqs):
        order, modes = shd.shd_num_degrees_vector(max_orders[f])
        Y_f = spspec.sph_harm(modes[None,:], order[None,:], angles[:,0:1], angles[:,1:2])
        B_f = spspec.spherical_jn(order[None,:], k[f]*r[:,None])

       # D_f = spspec.spherical_jn(order, k[f]*r_max)
        S_f = phi[f,:]

        Sigma_f = S_f[:,None] * Y_f * B_f #/ D_f[None,:]
        Sigma.append(Sigma_f)

    # for f in np.flip(np.arange(1, num_real_freqs)):
    #     order, modes = shd.shd_num_degrees_vector(max_orders[f])
    #     Y_f = spspec.sph_harm(modes[None,:], order[None,:], angles[:,0:1], angles[:,1:2])
    #     B_f = spspec.spherical_jn(order[None,:], k[f]*r[:,None])
    #     S_f = phi[f,:]
    #     Sigma_f = S_f[:,None] * Y_f * B_f
    #     Sigma.append(np.conj(Sigma_f))

    Sigma = np.concatenate(Sigma, axis=-1)
    system_mat = Sigma.conj().T @ Sigma + reg_param * np.eye(Sigma.shape[-1])
    print(f"Size of spatial spectrum system matrix: {system_mat.shape}")
    a = splin.solve(system_mat, Sigma.conj().T @ p, assume_a="pos")
    #a, residue, rank, singular_values = np.linalg.lstsq(Sigma, p, rcond=None)
    #a = lsqL2(Sigma, p, 1e-6)


    # ======= Reconstruction of RIR =======
    rir_est = np.zeros((num_real_freqs, num_eval), dtype=complex)
    r_eval, angles_eval = util.cart2spherical(pos_eval)
    ord_idx = 0
    for f in range(num_real_freqs):
        order, modes = shd.shd_num_degrees_vector(max_orders[f])
        num_ord = len(order)
        
        #j_denom = spspec.spherical_jn(order, k[f]*r_max)
        j_num = spspec.spherical_jn(order[None,:], k[f]*r_eval[:,None])

        Y = spspec.sph_harm(modes[None,:], order[None,:], angles_eval[:,0:1], angles_eval[:,1:2])
        rir_est[f, :] = np.sum(a[None,ord_idx:ord_idx+num_ord] * Y * j_num  , axis=-1) # / j_denom[None,:]
        ord_idx += num_ord

    if verbose:
        diagnostics = {}
        diagnostics["condition number"] = np.linalg.cond(Sigma).tolist()
        diagnostics["smallest singular value"] = splin.svdvals(Sigma)[0].tolist()
        diagnostics["largest singular"] = splin.svdvals(Sigma)[-1].tolist()
        diagnostics["r_max"] = r_max
        return rir_est, diagnostics
    return rir_est


def _sequence_stft_multiperiod(sequence, num_periods):
    """
    Assumes that the sequence is periodic.
    Assumes that sequence argument only contains one period
    
    Parameters
    ----------
    sequence : ndarray of shape (seq_len,)
    num_periods : int

    Returns
    -------
    Phi : ndarray of shape (seq_len, num_periods*seq_len)
    """
    Phi = _sequence_stft(sequence)
    return np.tile(Phi, (1, num_periods))

def _sequence_stft(sequence):
    """Might not correspond to the definition in the paper

    Parameters
    ----------
    sequence : ndarray of shape (seq_len,)

    Assume the sequence is periodic with period B

    Returns
    -------
    Phi : ndarray of shape (seq_len, seq_len)
        first axis contains frequency bins
        second axis contains time indices
    
    """
    if sequence.ndim == 2:
        sequence = np.squeeze(sequence, axis=0)
    assert sequence.ndim == 1
    B = sequence.shape[0]

    Phi = np.zeros((B, B), dtype=complex)

    for n in range(B):
        seq_vec = np.roll(sequence, -n) #so that n is the first element
        seq_vec = np.roll(seq_vec, -1) # so that n ends up last
        seq_vec = np.flip(seq_vec) # so that we get n first and then n-i as we move later in the vector
        for f in range(B):
            exp_vec = ft.idft_vector(f, B)
            Phi[f,n] = np.sum(exp_vec * seq_vec) 
    # Fast version, not sure if correct
    # for n in range(B):
    #     Phi[:,n] = np.fft.fft(np.roll(sequence, -n), axis=-1)
    return Phi














# ALL BELOW WAS PREVIOUSLY IN MOVINGMICROPHONE.PY









def inf_dimensional_shd_dynamic(p, pos, pos_eval, sequence, samplerate, c, reg_param, dir_coeffs, verbose=False):
    """
    Estimates the RIR at evaluation positions using data from a moving microphone
    using Bayesian inference of an infinite sequence of spherical harmonics

    Implements the method in J. Brunnström, M.B. Moeller, M. Moonen, 
    "Bayesian sound field estimation using moving microphones" 

    Assumptions:
    The noise covariance is a scaled identity matrix
    The data is measured over an integer number of periods of the sequence
    N = seq_len * M, where M is the number of periods that was measured
    The length of sequence is the length of the estimated RIR

    Parameters
    ----------
    p : ndarray of shape (N)
        sound pressure for each sample of the moving microphone
    pos : ndarray of shape (N, 3)
        position of the trajectory for each sample
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    sequence : ndarray of shape (seq_len) or (1, seq_len)
        the training signal used for the measurements
    samplerate : int
    c : float
        speed of sound
    reg_param : float
        regularization parameter
    dir_coeffs : ndarray of shape (N, num_coeffs)
        harmonic coefficients of microphone directivity, 
        Note that a higher number of coefficients will drastically increase the computational cost
    verbose : bool, optional
        if True, returns diagnostics, by default False

    Returns
    -------
    shd_coeffs : ndarray of shape (num_real_freqs, num_eval)
        time-domain harmonic coefficients of the estimated sound field
    """
    # ======= Argument parsing and constants =======
    if p.ndim >= 2:
        p = np.squeeze(p)
    N = p.shape[0]

    if sequence.ndim == 2:
        sequence = np.squeeze(sequence, axis=0)
    assert sequence.ndim == 1
    seq_len = sequence.shape[0]
    assert seq_len % 2 == 0 #Calculations later assume seq_len is even to get the Nyquist frequency
    num_periods = N // seq_len
    assert N % seq_len == 0

    wave_num = ft.get_wavenum(seq_len, samplerate, c)
    num_real_freqs = len(ft.get_real_freqs(seq_len, samplerate))

    assert pos.shape == (N, 3)
    assert pos_eval.ndim == 2 and pos_eval.shape[1] == 3

    if dir_coeffs.ndim == 2:
        dir_coeffs = dir_coeffs[None,:,:] #add a dimension for the number of frequencies
    assert dir_coeffs.ndim == 3
    assert dir_coeffs.shape[1] == N or dir_coeffs.shape[1] == 1
    assert dir_coeffs.shape[0] == num_real_freqs or dir_coeffs.shape[0] == 1
    

    # ======= Estimation of spherical harmonic coefficients =======
    Phi = _sequence_stft_multiperiod(sequence, num_periods)
    
    psi = calculate_psi(pos, dir_coeffs, wave_num, Phi, seq_len, num_real_freqs)
    noise_cov = reg_param * np.eye(N)
    regressor = splin.solve(psi + noise_cov, p, assume_a = "pos")
    regressor = Phi.conj()[:num_real_freqs,:] * regressor[None,:]

    est_sound_pressure = estimate_from_regressor(regressor, pos, pos_eval, wave_num[:num_real_freqs], dir_coeffs)
    
    if verbose:
        return est_sound_pressure, regressor, psi
    return est_sound_pressure


def estimate_from_regressor(regressor, pos, pos_eval, wave_num, dir_coeffs = None):
    """Takes the regressor from inf_dimensional_shd_dynamic, and gives back a sound field estimate. 

    Gives the same result as inf_dimensional_shd_dynamic, but is much faster since computing the regressor is the primary 
    computational cost. 
    Implements the method in J. Brunnström, M.B. Moeller, M. Moonen, "Bayesian sound field estimation using moving microphones" 


    Parameters
    ----------
    regressor : ndarray of shape (num_freqs, N)
        The regressor calculated by inf_dimensional_shd_dynamic. 
        Represents Phi* v in eq (31) from [brunnstromBayesian2024]
    pos : ndarray of shape (N, 3)
        position of the trajectory for each sample
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    wave_num : ndarray of shape (num_real_freq, )
        wavenumber, defined as w / c where w is the angular frequency
    dir_coeffs : ndarray of shape (N, num_coeffs), optional
        harmonic coefficients of microphone directivity, 
        Note that a higher number of coefficients will drastically increase the computational cost
        If not provided, the microphones are assumed to be omnidirectional.

    Returns
    -------
    shd_coeffs : ndarray of shape (num_real_freqs, num_eval)
        time-domain harmonic coefficients of the estimated sound field

    Notes
    -----
    Assumptions:
    The data is measured over an integer number of periods of the sequence
    N = seq_len * M, where M is the number of periods that was measured
    The length of sequence is the length of the estimated RIR
    """
    # ======= Argument parsing and constants =======
    N = pos.shape[0]
    assert pos.shape == (N, 3)
    assert pos_eval.ndim == 2 and pos_eval.shape[1] == 3

    num_real_freqs = wave_num.shape[-1]

    if dir_coeffs is None:
        return _estimate_from_regressor_omni(regressor, pos_eval, pos, wave_num)
        #dir_coeffs = shd.directivity_omni() * np.ones((N, 1))
    if dir_coeffs.ndim == 2:
        dir_coeffs = dir_coeffs[None,:,:] #add a dimension for the number of frequencies
    assert dir_coeffs.ndim == 3
    assert dir_coeffs.shape[1] == N or dir_coeffs.shape[1] == 1
    assert dir_coeffs.shape[0] == num_real_freqs or dir_coeffs.shape[0] == 1

    # ======= Reconstruction of RIR =======
    num_eval = pos_eval.shape[0]
    dir_omni = shd.directivity_omni() #* np.ones((num_eval, 1))
    dir_omni = dir_omni[None,:,:] # add a dimension for the number of frequencies

    # kernel_val = shd.translated_inner_product(pos_eval, pos, dir_omni, dir_coeffs, wave_num)
    # est_sound_pressure = np.squeeze(kernel_val @ regressor[:,:,None], axis=-1)

    num_eval = pos_eval.shape[0]
    est_sound_pressure = np.zeros(((num_real_freqs, num_eval)), dtype=complex)
    for i in range(num_eval):
        kernel_val = shd.translated_inner_product(pos_eval[i:i+1,:], pos, dir_omni, dir_coeffs, wave_num)
        est_sound_pressure[:,i:i+1] = np.squeeze(kernel_val @ regressor[:,:,None], axis=-1)
    return est_sound_pressure

def _estimate_from_regressor_omni(regressor, pos_eval, pos, k):
    """Takes the regressor from inf_dimensional_shd_dynamic, and gives back a sound field estimate. 
    Reconstructs the sound field at the evaluation points using the regressor matrix
    from est_inf_dimensional_shd_dynamic

    Parameters
    ----------
    regressor : ndarray of shape (num_real_freqs, N)
        regressor matrix from est_inf_dimensional_shd_dynamic
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    pos : ndarray of shape (N, 3)
        positions of the trajectory for each sample
    k : ndarray of shape (num_freq)
        wavenumbers

    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        estimated RIR per frequency at the evaluation points
    """
    num_real_freqs = regressor.shape[0]
    assert k.shape[-1] == num_real_freqs
    assert pos.shape[0] == regressor.shape[-1]

    est_sound_pressure = np.zeros((num_real_freqs, pos_eval.shape[0]), dtype=complex)
    for f in range(num_real_freqs):
        kernel_val = ki.kernel_helmholtz_3d(pos_eval, pos, k[f:f+1]).astype(complex)[0,:,:]
        est_sound_pressure[f, :] = np.sum(kernel_val * regressor[f,None,:], axis=-1)
    return est_sound_pressure



def calculate_psi(pos, dir_coeffs, k, Phi, seq_len, num_real_freqs):
    N = pos.shape[0]
    psi = np.zeros((N, N), dtype = float)

    print(f"starting TIP")
    # tip = shd.translated_inner_product(pos, pos, dir_coeffs, dir_coeffs, k[1:num_real_freqs-1])
    # phi_rank1_matrix = Phi[1:num_real_freqs-1,:,None] * Phi[1:num_real_freqs-1,None,:].conj()
    # psi = 2 * np.sum(np.real(tip * phi_rank1_matrix), axis=0)

    for f in range(1, num_real_freqs-1):
        print(f"Frequency {f}, going from 1 to {num_real_freqs-2} (inclusive)")
        phi_rank1_matrix = Phi[f,:,None] * Phi[f,None,:].conj()

        psi_f = np.squeeze(shd.translated_inner_product(pos, pos, dir_coeffs, dir_coeffs, k[f:f+1]), axis=0) * phi_rank1_matrix
        psi += 2 * np.real(psi_f)

    # no conjugation required for zeroth frequency and the Nyquist frequency, 
    # since they will be real already for a real input sequence
    psi += np.squeeze(np.real(shd.translated_inner_product(pos, pos, dir_coeffs, dir_coeffs, k[0:1])), axis=0) * np.real_if_close(Phi[0,:,None] * Phi[0,None,:])
    psi += np.squeeze(np.real(shd.translated_inner_product(pos, pos, dir_coeffs, dir_coeffs, k[seq_len//2:seq_len//2+1])), axis=0) * np.real_if_close(Phi[seq_len//2,:,None] * Phi[seq_len//2,None,:])
    return psi


def inf_dimensional_shd_dynamic_compiled(p, pos, pos_eval, sequence, samplerate, c, reg_param, dir_coeffs, verbose=False):
    """
    Estimates the RIR at evaluation positions using data from a moving microphone
    using Bayesian inference of an infinite sequence of spherical harmonics

    Implements the method in J. Brunnström, M.B. Moeller, M. Moonen, 
    "Bayesian sound field estimation using moving microphones" 

    Assumptions:
    The noise covariance is a scaled identity matrix
    The data is measured over an integer number of periods of the sequence
    N = seq_len * M, where M is the number of periods that was measured
    The length of sequence is the length of the estimated RIR

    Parameters
    ----------
    p : ndarray of shape (N)
        sound pressure for each sample of the moving microphone
    pos : ndarray of shape (N, 3)
        position of the trajectory for each sample
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    sequence : ndarray of shape (seq_len) or (1, seq_len)
        the training signal used for the measurements
    samplerate : int
    c : float
        speed of sound
    reg_param : float
        regularization parameter
    dir_coeffs : ndarray of shape (N, num_coeffs)
        harmonic coefficients of microphone directivity, 
        Note that a higher number of coefficients will drastically increase the computational cost
    verbose : bool, optional
        if True, returns diagnostics, by default False

    Returns
    -------
    shd_coeffs : ndarray of shape (num_real_freqs, num_eval)
        time-domain harmonic coefficients of the estimated sound field
    """
    # ======= Argument parsing and constants =======
    if p.ndim >= 2:
        p = np.squeeze(p)
    N = p.shape[0]

    if sequence.ndim == 2:
        sequence = np.squeeze(sequence, axis=0)
    assert sequence.ndim == 1
    seq_len = sequence.shape[0]
    assert seq_len % 2 == 0 #Calculations later assume seq_len is even to get the Nyquist frequency
    num_periods = N // seq_len
    assert N % seq_len == 0

    wave_num = ft.get_real_wavenum(seq_len, samplerate, c)
    num_real_freqs = wave_num.shape[-1]
    #len(ft.get_real_freqs(seq_len, samplerate))

    assert pos.shape == (N, 3)
    assert pos_eval.ndim == 2 and pos_eval.shape[1] == 3

    if dir_coeffs.ndim == 2:
        dir_coeffs = dir_coeffs[None,:,:] #add a dimension for the number of frequencies
    assert dir_coeffs.ndim == 3
    assert dir_coeffs.shape[1] == N or dir_coeffs.shape[1] == 1
    assert dir_coeffs.shape[0] == num_real_freqs or dir_coeffs.shape[0] == 1
    
    # ======= Estimation of spherical harmonic coefficients =======
    Phi = _sequence_stft_multiperiod(sequence, num_periods)
    Phi = Phi[:num_real_freqs,:]
    
    psi = calculate_psi_compiled(pos, dir_coeffs, wave_num, Phi, seq_len, num_real_freqs)
    psi = np.asarray(psi) # convert to numpy array, as jax is no longer necessary
    noise_cov = reg_param * np.eye(N)
    psi_plus_noise_cov = psi + noise_cov
    try:
        regressor = splin.solve(psi_plus_noise_cov, p, assume_a = "pos")
    except np.linalg.LinAlgError:
        print(f"LinAlgError at reg_param = {reg_param}. Singular psi (plus noise cov) matrix")
        print(f"Calculating least-squares solution instead")
        regressor = splin.lstsq(psi_plus_noise_cov, p)[0]
    regressor = Phi.conj() * regressor[None,:]

    if verbose:
        print(f"Computing eval estimates from regressor")
    est_sound_pressure = estimate_from_regressor_compiled(regressor, pos, pos_eval, wave_num, dir_coeffs)

    if verbose:
        return est_sound_pressure, regressor, psi
    return est_sound_pressure


def estimate_from_regressor_compiled(regressor, pos, pos_eval, wave_num, dir_coeffs = None):
    """Takes the regressor from inf_dimensional_shd_dynamic, and gives back a sound field estimate. 

    Gives the same result as inf_dimensional_shd_dynamic, but is much faster since computing the regressor is the primary 
    computational cost. 
    Implements the method in J. Brunnström, M.B. Moeller, M. Moonen, "Bayesian sound field estimation using moving microphones" 


    Parameters
    ----------
    regressor : ndarray of shape (num_real_freqs, N)
        The regressor calculated by inf_dimensional_shd_dynamic. 
        Represents Phi* v in eq (31) from [brunnstromBayesian2024]
    pos : ndarray of shape (N, 3)
        position of the trajectory for each sample
    pos_eval : ndarray of shape (num_eval, 3)
        positions of the evaluation points
    samplerate : int
    c : float
        speed of sound
    dir_coeffs : ndarray of shape (N, num_coeffs), optional
        harmonic coefficients of microphone directivity, 
        Note that a higher number of coefficients will drastically increase the computational cost
        If not provided, the microphones are assumed to be omnidirectional.

    Returns
    -------
    est_sound_pressure : ndarray of shape (num_real_freqs, num_eval)
        time-domain harmonic coefficients of the estimated sound field

    Notes
    -----
    Assumptions:
    The data is measured over an integer number of periods of the sequence
    N = seq_len * M, where M is the number of periods that was measured
    The length of sequence is the length of the estimated RIR
    """
    # ======= Argument parsing and constants =======
    N = pos.shape[0]
    assert pos.shape == (N, 3)
    assert pos_eval.ndim == 2 and pos_eval.shape[1] == 3

    num_real_freqs = wave_num.shape[-1]
    assert wave_num.ndim == 1
    assert regressor.shape == (num_real_freqs, N)

    if dir_coeffs is None:
        dir_coeffs = shd.directivity_omni() * np.ones((N, 1))
    if dir_coeffs.ndim == 2:
        dir_coeffs = dir_coeffs[None,:,:] #add a dimension for the number of frequencies
    assert dir_coeffs.ndim == 3
    assert dir_coeffs.shape[1] == N or dir_coeffs.shape[1] == 1
    assert dir_coeffs.shape[0] == num_real_freqs or dir_coeffs.shape[0] == 1

    if dir_coeffs.shape[0] != 1:
        raise NotImplementedError("Only implemented for frequency-independent directivity coefficients")
    #dir_coeffs = dir_coeffs[0,:,:]

    # ======= Reconstruction of RIR =======
    num_eval = pos_eval.shape[0]
    dir_omni = shd.directivity_omni() #* np.ones((num_eval, 1))
    dir_omni = dir_omni[None,:,:] # add a dimension for the number of frequencies

    num_eval = pos_eval.shape[0]
    est_sound_pressure = np.zeros(((num_real_freqs, num_eval)), dtype=complex)
    for i in range(num_eval):
        #if i % 10 == 0:
        print(f"Estimating sound pressure at position {i} of {num_eval}")
        kernel_val = shd.translated_inner_product(pos_eval[i:i+1,:], pos, dir_omni, dir_coeffs, wave_num)
        est_sound_pressure[:,i:i+1] = np.squeeze(kernel_val @ regressor[:,:,None], axis=-1)

    # MAKE A TRANSLATION OPERATOR THAT WORKS FOR MIXED 0TH AND 1ST ORDERS. THEN IMPLEMENT FOLLOWING
    #max_order = shd.shd_max_order(dir_coeffs.shape[-1])
    #gaunt_set = _calculate_gaunt_set(0, max_order)
    #return _estimate_from_regressor_compiled(regressor, pos, pos_eval, wave_num, dir_omni, dir_coeffs, gaunt_set)
    return est_sound_pressure

def _estimate_from_regressor_compiled(regressor, pos, pos_eval, wave_num, dir_omni, dir_coeffs, gaunt_set):

    def psi_scan_loop(carry, arg_slice):
        (pos_i, dir_coeffs_i) = arg_slice

        def psi_scan_inner_loop(carry_inner, arg_slice_inner):
            (pos_j, dir_coeffs_j) = arg_slice_inner

            pos_diff = pos_i[None,:] - pos_j[None,:]

            T = translation_operator(pos_diff, wave_num, gaunt_set)
            inner_product = jnp.moveaxis(jnp.conj(dir_coeffs_i)[None,None,:,None], -1, -2) @ T @ dir_coeffs_j[None,None,:,None]
            inner_product = jnp.squeeze(inner_product)

            psi_ij = 2 * jnp.sum(jnp.real(inner_product[1:-1,...]))
            psi_ij = psi_ij + jnp.real(inner_product[0,...]) + jnp.real(inner_product[-1,...])

            return carry_inner, psi_ij
        
        _, psi_i = jax.lax.scan(psi_scan_inner_loop, 0, (pos, dir_coeffs))        
        return (carry, psi_i)
    
    _, psi = jax.lax.scan(psi_scan_loop, 0, (pos_eval, dir_omni))


    kernel_val = shd.translated_inner_product(pos_eval, pos, dir_omni, dir_coeffs, wave_num)
    est_sound_pressure = np.squeeze(kernel_val @ regressor[:,:,None], axis=-1)
    return est_sound_pressure

def _calculate_psi_compiled_each_freq(pos, dir_coeffs, k, Phi, seq_len, num_real_freqs, gaunt_set):
    N = pos.shape[0]
    psi = np.zeros((N, N), dtype = float)

    for f in range(1, num_real_freqs-1):
        print(f"Frequency {f}, going from 1 to {num_real_freqs-2} (inclusive)")
        phi_rank1_matrix = Phi[f,:,None] * Phi[f,None,:].conj()

        psi_f = np.squeeze(translated_inner_product(pos, dir_coeffs, k[f:f+1], gaunt_set), axis=0) * phi_rank1_matrix
        psi += 2 * np.real(psi_f)

    # no conjugation required for zeroth frequency and the Nyquist frequency, 
    # since they will be real already for a real input sequence
    psi += np.squeeze(np.real_if_close(translated_inner_product(pos, dir_coeffs, k[0:1], gaunt_set)), axis=0) * np.real_if_close(Phi[0,:,None] * Phi[0,None,:])
    psi += np.squeeze(np.real(translated_inner_product(pos, dir_coeffs, k[seq_len//2:seq_len//2+1], gaunt_set)), axis=0) * np.real_if_close(Phi[seq_len//2,:,None] * Phi[seq_len//2,None,:])
    return psi


def calculate_psi_compiled(pos, dir_coeffs, wave_num, Phi, seq_len, num_real_freqs):
    if dir_coeffs.ndim == 2:
        dir_coeffs = dir_coeffs[:,None,:]
    assert pos.ndim == 2
    assert pos.shape[1] == 3
    num_pos = pos.shape[0]

    assert wave_num.ndim == 1
    assert wave_num.shape[0] == num_real_freqs
    
    assert dir_coeffs.ndim == 3
    if dir_coeffs.shape[0] == wave_num.shape[0]:
        raise NotImplementedError("Only implemented for frequency-independent directivity coefficients")
    assert dir_coeffs.shape[0] == 1
    assert dir_coeffs.shape[1] == pos.shape[0] or dir_coeffs.shape[1] == 1

    assert Phi.ndim == 2
    assert Phi.shape == (num_real_freqs, num_pos)

    max_order = 1
    gaunt_set = _calculate_gaunt_set(max_order, max_order)

    print("Running compiled Psi calculation")
    psi = _calculate_psi_compiled(pos, dir_coeffs, wave_num, Phi, gaunt_set)
    return psi


@jax.jit
def _calculate_psi_compiled(pos, dir_coeffs, wave_num, Phi, gaunt_set):
    num_pos = pos.shape[0]

    Phi = Phi.T # we need to scan over leading axis
    dir_coeffs = dir_coeffs[0,:,:] # assume frequency independent directivity

    def psi_scan_loop(carry, arg_slice):
        (pos_i, phi_i, dir_coeffs_i) = arg_slice

        def psi_scan_inner_loop(carry_inner, arg_slice_inner):
            (pos_j, phi_j, dir_coeffs_j) = arg_slice_inner

            pos_diff = pos_i[None,:] - pos_j[None,:]
            phi_factor = phi_i * jnp.conj(phi_j)

            T = translation_operator(pos_diff, wave_num, gaunt_set)
            inner_product = jnp.moveaxis(jnp.conj(dir_coeffs_i)[None,None,:,None], -1, -2) @ T @ dir_coeffs_j[None,None,:,None]
            inner_product = jnp.squeeze(inner_product) * phi_factor

            psi_ij = 2 * jnp.sum(jnp.real(inner_product[1:-1,...]))
            psi_ij = psi_ij + jnp.real(inner_product[0,...]) + jnp.real(inner_product[-1,...])

            return carry_inner, psi_ij
        _, psi_i = jax.lax.scan(psi_scan_inner_loop, 0, (pos, Phi, dir_coeffs))        
        return (carry, psi_i)
    _, psi = jax.lax.scan(psi_scan_loop, 0, (pos, Phi, dir_coeffs))
    return psi

@jax.jit
def _calculate_psi_compiled_single_loop_in_scan(pos, dir_coeffs, wave_num, Phi, gaunt_set):
    num_pos = pos.shape[0]

    Phi = Phi.T # we need to scan over leading axis
    dir_coeffs = dir_coeffs[0,:,:] # assume frequency independent directivity

    def psi_scan_loop(carry, arg_slice):
        (pos_i, phi_i, dir_coeffs_i) = arg_slice

        psi_vals = []
        for j in range(num_pos):
            pos_diff = pos_i[None,:] - pos[j:j+1,:]
            phi_factor = phi_i * jnp.conj(Phi[j,:])

            T = translation_operator(pos_diff, wave_num, gaunt_set)
            inner_product = jnp.moveaxis(jnp.conj(dir_coeffs_i)[None,None,:,None], -1, -2) @ T @ dir_coeffs[None,j:j+1,:,None]
            inner_product = jnp.squeeze(inner_product) * phi_factor

            psi_val = 2 * jnp.sum(jnp.real(inner_product[1:-1,...]))
            psi_val = psi_val + jnp.real(inner_product[0,...]) + jnp.real(inner_product[-1,...])
            psi_vals.append(psi_val)
        psi_i = jnp.stack(psi_vals, axis=0)
            
        carry = carry + 1
        return (carry, psi_i)
    
    carry, psi = jax.lax.scan(psi_scan_loop, 0, (pos, Phi, dir_coeffs))
    return psi


@jax.jit
def _calculate_psi_compiled_with_python_loops(pos, dir_coeffs, wave_num, Phi, gaunt_set):
    num_pos = pos.shape[0]

    psi_vals = []
    for i in range(num_pos):
        for j in range(num_pos):
            pos_diff = pos[i:i+1,:] - pos[j:j+1,:]
            phi_factor = Phi[:,i] * jnp.conj(Phi[:,j])

            T = translation_operator(pos_diff, wave_num, gaunt_set)
            inner_product = jnp.moveaxis(jnp.conj(dir_coeffs)[:,i:i+1,:,None], -1, -2) @ T @ dir_coeffs[:,j:j+1,:,None]
            inner_product = jnp.squeeze(inner_product) * phi_factor

            psi_val = 2 * jnp.sum(jnp.real(inner_product[1:-1,...]))
            psi_val = psi_val + jnp.real(inner_product[0,...]) + jnp.real(inner_product[-1,...])
            psi_vals.append(psi_val)

    psi = jnp.stack(psi_vals, axis=0)
    psi = jnp.reshape(psi, (num_pos, num_pos))
    return psi


@jax.jit
def translated_inner_product(pos, dir_coeffs, wave_num, gaunt_set):
    """
    
    <T(r_1 - r_2, omega_m) gamma_2(omega_m), gamma_1(omega_m)>
    
    Parameters
    ----------
    pos1 : ndarray of shape (num_pos1, 3)
        positions of the first set of measurement points
    pos2 : ndarray of shape (num_pos2, 3)
        positions of the second set of measurement points
    dir_coeffs1 : ndarray of shape (num_freqs, num_coeffs1) or (num_freqs, num_pos1, num_coeffs1)
        coefficients of the directivity function for the first set of measurement points
    dir_coeffs2 : ndarray of shape (num_freqs, num_coeffs2) or (num_freqs, num_pos1, num_coeffs1)
        coefficients of the directivity function for the second set of mea surement points
    wave_num : ndarray of shape (num_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.

    Returns
    -------
    psi : ndarray of shape (num_freqs, num_pos1, num_pos2)
        inner product of the translated directivity functions
    """
    assert pos.ndim == 2
    assert pos.shape[-1] == 3
    num_pos = pos.shape[0]

    assert wave_num.ndim == 1
    num_freqs = wave_num.shape[0]
    
    if dir_coeffs.ndim == 2:
        dir_coeffs = dir_coeffs[:,None,:]
    assert dir_coeffs.ndim == 3
    assert dir_coeffs.shape[1] == num_pos or dir_coeffs.shape[1] == 1
    assert dir_coeffs.shape[0] == num_freqs or dir_coeffs.shape[0] == 1

    pos_diff = pos[:,None,:] - pos[None,:,:]

    translated_coeffs2 = jax.vmap(translate_shd_coeffs, in_axes=(None, 0, None, None), out_axes=1)(dir_coeffs, pos_diff, wave_num, gaunt_set)
    inner_product_matrix = jnp.sum(translated_coeffs2 * dir_coeffs.conj()[:,:,None,:], axis=-1)
    return inner_product_matrix



@jax.jit
def translate_shd_coeffs(shd_coeffs, pos, wave_num, gaunt_set):
    """Translates the provided shd_coeffs to another expansion center
    
    shd_coeffs(pos_orig + pos) = translate(pos) shd_coeffs(pos_orig)
    
    Parameters
    ----------
    shd_coeffs : ndarray of shape (num_freqs, num_coeffs,) or (num_freqs, num_pos, num_coeffs,)
        coefficients of the sequence
    pos : ndarray of shape (num_pos, 3)
        position argument to the translation operator
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    max_order_output : int
        maximum order of the translated coefficients

    Returns
    -------
    shd_coeffs_translated : ndarray of shape (num_freqs, num_pos, num_coeffs,)
        translated coefficients
        if num_pos == 1, the returned array will have shape (num_freqs, num_coeffs,)
    """
    if shd_coeffs.ndim == 2:
        shd_coeffs = shd_coeffs[:,None,:]
    assert pos.ndim == 2
    assert pos.shape[1] == 3
    assert wave_num.ndim == 1
    assert shd_coeffs.ndim == 3
    assert shd_coeffs.shape[0] == wave_num.shape[0] or shd_coeffs.shape[0] == 1
    assert shd_coeffs.shape[1] == pos.shape[0] or shd_coeffs.shape[1] == 1
    #num_coeffs = shd_coeffs.shape[-1]

    #max_order_input = shd.shd_max_order(num_coeffs)

    T_all = translation_operator(pos, wave_num, gaunt_set)
    translated_coeffs = T_all @ shd_coeffs[...,None]

    num_pos = pos.shape[0]
    if num_pos == 1:
        translated_coeffs = jnp.squeeze(translated_coeffs, axis=1)
    return jnp.squeeze(translated_coeffs, axis=-1)

@jax.jit
def translation_operator(pos, wave_num, gaunt_set):
    """Translation operator for harmonic coefficients, such that 
    shd_coeffs(pos_orig + pos) = T(pos) @ shd_coeffs(pos_orig)

    Defined according to T = hat{S}^T, where hat{S} is the basis translation matrix defined in
    P. Martin, Multiple scattering: Interaction of time-harmonic waves with N obstacles.
    This definition is consistent with the definition (6) in Ueno et al., Sound Field Recording Using
    Distributed Microphones Based on Harmonic Analysis of Infinite Order. 
    
    Parameters
    ----------
    pos : ndarray of shape (num_pos, 3)
        position argument to the translation operator
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    max_order_input : int
        maximum order of the coefficients that should be translated
    max_order_output : int
        maximum order of the translated coefficients

    Returns
    -------
    T : ndarray of shape (num_freqs, num_pos, num_coeffs_output, num_coeffs_input)
        translation operator, such that shd_coeffs(pos_orig + pos) = T(pos) @ shd_coeffs(pos_orig)
    """

    S = basis_translation_3_80(pos, wave_num, gaunt_set)
    T = jnp.moveaxis(S, -1, -2)
    return T


@jax.jit
def basis_translation_3_80(pos, wave_num, gaunt_set):
    """Translation operator for shd basis function, such that 
    shd_basis(pos_orig + pos) = T(pos) @ shd_basis(pos_orig)

    Implemented according to equation 3.80 in 
    P. Martin, Multiple scattering: Interaction of time-harmonic waves with N obstacles.

    Hardcoded for max_order = 1 to make compilation feasible
    Currently returns nan when pos is (0,0,0)
    
    Parameters
    ----------
    pos : ndarray of shape (num_pos, 3)
        position argument to the translation operator
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    max_order_input : int
        maximum order of the coefficients that should be translated
    max_order_output : int
        maximum order of the translated coefficients

    Returns
    -------
    T : ndarray of shape (num_freqs, num_pos, num_coeffs_output, num_coeffs_input)
        translation operator, such that shd_basis(pos_orig + pos) = T(pos) @ shd_basis(pos_orig)
    """
    max_order = 1
    num_coeffs = 4
    num_pos = pos.shape[0]
    num_freq = wave_num.shape[0]

    tr_op = jnp.zeros((num_freq, num_pos, num_coeffs, num_coeffs), dtype=complex)
    orders = np.array([0, 1, 1, 1])
    degrees = np.array([0, -1, 0, 1])

    max_tot_order = 2 * max_order 
    num_coeffs_tot = 9
    orders_tot = np.array([0,  1,  1,  1,  2,  2,  2,  2,  2])
    degrees_tot = np.array([0, -1,  0,  1, -2, -1,  0,  1,  2])

    num_q = max_tot_order + 1
    all_q = np.arange(max_tot_order+1)
    q_lst = []
    for out_idx, (n, m) in enumerate(zip(orders, degrees)):
        q_lst.append([])
        for in_idx, (nu, mu) in enumerate(zip(orders, degrees)):
            q_lst[out_idx].append(np.arange(np.abs(mu-m), n+nu+1))

    harm_idx_lst = []
    for out_idx, (n, m) in enumerate(zip(orders, degrees)):
        harm_idx_lst.append([])
        for in_idx, (nu, mu) in enumerate(zip(orders, degrees)):
            harm_idx_lst[out_idx].append(shd_coeffs_order_degree_to_index(q_lst[out_idx][in_idx], mu-m))

    radius, azimuth, zenith = cart2spherical(pos)
    radius = radius[None, None, ...]
    azimuth = azimuth[None, ...]
    zenith = zenith[None, ...]

    orders_tot = orders_tot[:, None]
    degrees_tot = degrees_tot[:, None]

    wave_num = wave_num[:, None, None]
    all_q = all_q[None,:,None]

    bessel_shape = (num_freq, num_q, num_pos)
    bessel = spherical_jn(jnp.ravel(jnp.broadcast_to(all_q, bessel_shape)), 
                                        jnp.ravel(jnp.broadcast_to(wave_num * radius, bessel_shape))) # ordered as (num_freq, num_q, num_pos)
    bessel = jnp.reshape(bessel, bessel_shape)

    harm_shape = (num_coeffs_tot, num_pos)
    harm = jax.scipy.special.sph_harm(jnp.ravel(jnp.broadcast_to(degrees_tot, harm_shape)), 
                                        jnp.ravel(jnp.broadcast_to(orders_tot, harm_shape)), 
                                        jnp.ravel(jnp.broadcast_to(azimuth, harm_shape)), 
                                        jnp.ravel(jnp.broadcast_to(zenith, harm_shape)), 
                                        max_tot_order) # ordered as (num_coeffs_tot, num_pos)
    harm = jnp.reshape(harm, harm_shape)
    harm = jnp.conj(harm)

    for out_idx, (n, m) in enumerate(zip(orders, degrees)):
        for in_idx, (nu, mu) in enumerate(zip(orders, degrees)):
            q_array = q_lst[out_idx][in_idx]
            harm_idxs = harm_idx_lst[out_idx][in_idx]

            basis_val = bessel[:,q_array,:] * harm[None,harm_idxs, :]
            g = (1j)**q_array[None,:,None] * gaunt_set[out_idx, in_idx, q_array][None,:,None]
            factor = 4*jnp.pi * (1j)**(nu-n) * (-1.0)**(m)
            sum_val = factor * jnp.sum(basis_val * g, axis=1)
            tr_op = tr_op.at[..., out_idx, in_idx].set(sum_val)


    zero_pos = jnp.linalg.norm(pos, axis=-1) == 0
    tr_op = jnp.where(zero_pos[None,:,None,None], jnp.eye(num_coeffs, num_coeffs)[None,None,:,:], tr_op)
    return tr_op



def _calculate_gaunt_set(max_order_input, max_order_output):
    """Numpy function for precaculating the Gaunt coefficients for the translation operator.

    Used to avoid having to write a jax implementation of wigner coefficients. 
    """
    num_coeffs_input = shd_num_coeffs(max_order_input)
    num_coeffs_output = shd_num_coeffs(max_order_output)
    orders_input, degrees_input = shd_num_degrees_vector(max_order_input)
    orders_output, degrees_output = shd_num_degrees_vector(max_order_output)
    gaunt = np.zeros((num_coeffs_output, num_coeffs_input, max_order_input+max_order_output+1), dtype=float)
    for out_idx, (n, m) in enumerate(zip(orders_output, degrees_output)):
        for in_idx, (nu, mu) in enumerate(zip(orders_input, degrees_input)):
            for q in range(n+nu+1):
                if np.abs(mu-m) <= q:
                    g = shd.gaunt_coefficient(n, m, nu, -mu, q)
                    gaunt[out_idx, in_idx, q] = g
    return gaunt


def shd_coeffs_order_degree_to_index(order, degree):
    return order**2 + order + degree



def factorial(n):
    return jax.scipy.special.factorial(n, exact=False)

@jax.jit
def spherical_jn(v, z):
    """Calculates j_v(z), the spherical Bessel function of the first kind of order v

    Currently limited to orders 0, 1 and 2. See documentation for spherical_jn_max_order_2 for more information.

    Parameters
    ----------
    v : ndarray of shape (num_points,)
        order of the spherical Bessel function
    z : ndarray of shape (num_points,)
        argument of the spherical Bessel function

    Returns
    -------
    order_v : ndarray of shape (num_points,)
        spherical Bessel function of the first kind of order v
    """
    return spherical_jn_max_order_2(v, z)

@jax.jit
def spherical_jn_max_order_2(v, z):
    """Calculates j_v(z), the spherical Bessel function of the first kind of order v

    Hard codes the function for order 0, 1 and 2, which is the ones required for sound field
    estimation with first order directional microphones. Uses a small-argument approximation to maintain
    acceptable numerical stability for small arguments.

    Parameters
    ----------
    v : ndarray of shape (num_points,)
        order of the spherical Bessel function
    z : ndarray of shape (num_points,)
        argument of the spherical Bessel function

    Returns
    -------
    order_v : ndarray of shape (num_points,)
        spherical Bessel function of the first kind of order v
        
    Notes
    -----
    The definitions are written explicitly at https://mathworld.wolfram.com/SphericalBesselFunctionoftheFirstKind.html
    Alternatively, the second-order function can be derived from the recursion defined in (14) in https://www.fresco.org.uk/functions/barnett/APP23.pdf

    The small-argument approximation is written explicitly at the end of page 2 in
    https://www.damtp.cam.ac.uk/user/tong/aqm/bessel.pdf

    """
    order_0 = jnp.sinc(z / jnp.pi)
    order_1 = jnp.sinc(z / jnp.pi) / z - jnp.cos(z) / z

    div_term = 3 / z**2 
    order_2 = (div_term-1) * jnp.sinc(z / jnp.pi) - div_term * jnp.cos(z)

    order_1_small = z / 3
    order_2_small = z**2 / 15

    ORDER1_SWITCH_VALUE = 2.5e-4
    ORDER2_SWITCH_VALUE = 6e-3

    order_1 = jnp.where(z < ORDER1_SWITCH_VALUE, order_1_small, order_1)
    order_2 = jnp.where(z < ORDER2_SWITCH_VALUE, order_2_small, order_2)

    all_orders = jnp.stack([order_0, order_1, order_2], axis=0)
    return all_orders[v, jnp.arange(z.shape[-1])]


def _spherical_jn_recursion(v, z):
    """Uses a simple recursion to calculate the spherical Bessel function of the first kind

    Is numerically unstable for small arguments for orders over 2. 
    
    Parameters
    ----------
    v : array of shape (num_points,)
        order of the spherical Bessel function
    z : array of shape (num_points,)
        argument of the spherical Bessel function
    
    Returns
    -------
    order_v : array_like
        spherical Bessel function of the first kind of order v

    Notes
    -----
    Uses (14) from:
    https://www.fresco.org.uk/functions/barnett/APP23.pdf
    """
    max_v = jnp.max(v)
    all_jn = []

    order_v_minus2 = jnp.sin(z) / z
    order_v_minus1 = jnp.sin(z) / z**2 - jnp.cos(z) / z
    all_jn.append(order_v_minus2)
    all_jn.append(order_v_minus1)

    for i in range(2, max_v+1):
        order_v = (2*i + 1) / z * order_v_minus1 - order_v_minus2
        all_jn.append(order_v)
        order_v_minus2 = order_v_minus1
        order_v_minus1 = order_v
    all_jn = jnp.stack(all_jn, axis=0)

    return all_jn[v,np.arange(z.shape[-1])]

def _spherical_jn_recursion_scalar_order(v, z):
    """
    
    Parameters
    ----------
    v : int
        order of the spherical Bessel function
    z : array_like
        argument of the spherical Bessel function
    
    Returns
    -------
    order_v : array_like
        spherical Bessel function of the first kind of order v
    """
    order_v_minus2 = jnp.sin(z) / z
    if v == 0:
        return order_v_minus2
    order_v_minus1 = jnp.sin(z) / z**2 - jnp.cos(z) / z
    if v == 1:
        return order_v_minus1

    for i in range(2, v+1):
        order_v = (2*i + 1) / z * order_v_minus1 - order_v_minus2
        order_v_minus2 = order_v_minus1
        order_v_minus1 = order_v
    return order_v

def _spherical_jn_log_series(v, z, max_idx=10):
    """Calculates j_v(z), the spherical Bessel function of the first kind of order v

    only accepts integer order v and real arguments z

    Parameters
    ----------
    v : ndarray of shape (num_points,)
        order of the spherical Bessel function
    z : ndarray of shape (num_points,)
        argument of the spherical Bessel function
    max_idx : int
        maximum index of the series expansion

    uses series expansion
    
    Notes
    -----
    max_ind does not increase the accuracy much after 10 or 15. 
    """
    #max_idx = 35
    #output = jnp.zeros_like(z)
    v = v[:,None]
    z = z[:,None]
    k = jnp.arange(max_idx)[None,:]

    log_terms = v * jnp.log(2) + (v + 2*k) * jnp.log(z) + jax.scipy.special.gammaln(k + v + 1) - jax.scipy.special.gammaln(k + 1) - jax.scipy.special.gammaln(2*k + 2*v + 2)
    all_series_terms = (-1)**(k) * jnp.exp(log_terms)
    output = jnp.sum(all_series_terms, axis=-1)
    return output


def _spherical_jn_series(v, z, max_idx=10):
    """Calculates j_v(z), the spherical Bessel function of the first kind of order v

    only accepts integer order v and real arguments z

    Parameters
    ----------
    v : ndarray of shape (num_points,)
        order of the spherical Bessel function
    z : ndarray of shape (num_points,)
        argument of the spherical Bessel function
    max_idx : int
        maximum index of the series expansion

    uses series expansion
    
    Notes
    -----
    max_ind = 35 allows for high accuracy for z < 20
    requires 64bit precision, otherwise it will return NaNs. 
    up to 15 or 20 max_idx is generally fine if 32bit precision is used
    """
    #max_idx = 35
    #output = jnp.zeros_like(z)
    v = v[:,None]
    z = z[:,None]
    k = jnp.arange(max_idx)[None,:]

    order_factor = 2**(v) * (-1)**(k) * factorial(k + v) / (factorial(k) * factorial(2*k + 2*v + 1))
    argument = z**(v) * z**(2*k)

    return jnp.sum(order_factor * argument, axis=-1)

def cart2spherical(cart_coord):
    """Transforms the provided cartesian coordinates to spherical coordinates

    Parameters
    ----------
    cart_coord : ndarray of shape (num_points, 3)

    Returns
    -------
    r : ndarray of shape (num_points, 1)
        radius of each point
    angle : ndarray of shape (num_points, 2)
        angle[:,0] is theta, the angle in the xy plane, where 0 is x direction, pi/2 is y direction
        angle[:,1] is phi, the zenith angle, where 0 is z direction, pi is negative z direction
    """
    r = jnp.linalg.norm(cart_coord, axis=1)
    r_xy = jnp.linalg.norm(cart_coord[:,:2], axis=1)

    azimuth = jnp.arctan2(cart_coord[:,1], cart_coord[:,0])
    zenith = jnp.arctan2(r_xy, cart_coord[:,2])
    #angle = jnp.concatenate((theta[:,None], phi[:,None]), axis=1)
    return (r, azimuth, zenith)

def shd_basis(pos, order, degree, wave_num, max_order):
    """Spherical harmonic basis function for sound field in 3D

    Implements: sqrt(4pi) j_order(kr) Y_order^degree(polar angle, zenith angle)
    degree and order of the spherical harmonics might be swapped according to some
    definitions.
    
    Parameters
    ----------
    pos : Tensor of shape (num_pos, 3)
        positions of the evaluation points
    order : ndarray of shape (num_coeffs,)
        order of the spherical harmonics. Can be any non-negative integer
    degree : ndarray of shape (num_coeffs,)
        degree of the spherical harmonics. Must satisfy |degree[i]| <= order[i] for all i
    wave_num : float or ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular 
        frequency and c is the speed of sound
    max_order : int
        required for jax to be able to compile the function

    Returns
    -------
    f : ndarray of shape (num_freq, num_coeffs, num_pos)
        values of the spherical harmonic basis function at the evaluation points

    Notes
    -----
    See more detailed derivations in any of:
    [1] Natsuki Ueno, Shoichi Koyama, Hiroshi Saruwatai, 'Sound Field Recording Using 
    Distributed Microphones Based on Harmonic Analysis of Infinite Order'
    [2] J. Brunnström, M.B. Moeller, M. Moonen, 'Bayesian sound field estimation using
    moving microphones'
    """
    #Parse arguments to get correct shapes
    num_pos = pos.shape[0]
    num_coeffs = order.shape[0]
    num_freq = wave_num.shape[0]

    radius, theta, phi = cart2spherical(pos)

    size = (num_freq, num_coeffs, num_pos)
    radius = jnp.broadcast_to(radius[None,None,:], size)
    theta = jnp.broadcast_to(theta[None,None,:], size)
    phi = jnp.broadcast_to(phi[None,None,:], size)
    order = jnp.broadcast_to(order[None,:,None], size)
    degree = jnp.broadcast_to(degree[None,:,None], size)
    wave_num = jnp.broadcast_to(wave_num[:,None,None], size)

    radius = jnp.reshape(radius, -1)
    theta = jnp.reshape(theta, -1)
    phi = jnp.reshape(phi, -1)
    order = jnp.reshape(order, -1)
    degree = jnp.reshape(degree, -1)
    wave_num = jnp.reshape(wave_num, -1)

    harm = jax.scipy.special.sph_harm(degree, order, theta, phi, n_max = max_order)

    # Calculate the function values

    bessel = spherical_jn(order, wave_num * radius)
    f = jnp.sqrt(4*jnp.pi) * bessel * harm
    f = jnp.reshape(f, size)
    return f

def shd_num_coeffs(max_order):
    """Number of coefficients required to represent a spherical harmonic function of a given order
    
    Parameters
    ----------
    max_order : int
        maximum order of the spherical harmonics
    
    Returns
    -------
    num_coeffs : int
        number of coefficients required to represent a spherical harmonic function of the given order
    """
    return (max_order+1)**2

def shd_num_degrees_vector(max_order : int):
    """
    Constructs a vector with the index of each order and degree
    when order = n, the degrees are only non-zero for -n <= degree <= n

    Parameters
    ----------
    max_order : int
        is the maximum order that is included

    Returns
    -------
    order : ndarray of shape ()
    degree : ndarray of shape ()
    """
    order = []
    degree = []

    for n in range(max_order+1):
        pos_degrees = jnp.arange(n+1)
        degree_n = jnp.concatenate((-jnp.flip(pos_degrees[1:]), pos_degrees))
        degree.append(degree_n)

        order.append(n*jnp.ones_like(degree_n))
    degree = jnp.concatenate(degree)
    order = jnp.concatenate(order)
    return order, degree



import numpy as np
#from hypothesis import given
from pyinstrument import Profiler
import scipy.special as special

import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp


import aspcore.fouriertransform as ft
import aspcore.montecarlo as mc

import aspcol.sphericalharmonics as sph
import aspcol.soundfieldestimation as sfe
import aspcol.movingmicrophone as mm
import aspcol.plot as aspplot

import matplotlib.pyplot as plt


def _random_shd_coeffs(max_order, num_freqs, rng=None):
    if rng is None:
        rng = np.random.default_rng()

    shd_coeffs = rng.normal(size=(num_freqs, sph.shd_num_coeffs(max_order))) + \
        1j * rng.normal(size=(num_freqs, sph.shd_num_coeffs(max_order)))
    return shd_coeffs

def _random_mm_arguments(rng, seq_len, num_periods, max_order, num_eval):
    num_pos = seq_len * num_periods
    num_freqs = seq_len

    p = rng.normal(size=(1, num_pos))
    pos = rng.uniform(-1, 1, size=(num_pos, 3))
    pos_eval = rng.uniform(-1, 1, size=(num_eval, 3))
    dir_coeffs = np.stack([_random_shd_coeffs(max_order, 1, rng) for _ in range(num_pos)], axis=1)
    seq = rng.uniform(-1, 1, size=(seq_len,))
    reg_param = 10 ** rng.uniform(-6, 1)
    return p, pos, pos_eval, seq, reg_param, dir_coeffs

def _random_psi_arguments(rng, seq_len, num_periods, max_order, samplerate, c):
    num_pos = seq_len * num_periods
    num_freqs = seq_len

    pos = rng.uniform(-1, 1, size=(num_pos, 3))
    wave_num = ft.get_real_wavenum(num_freqs, samplerate, c)
    num_real_freqs = wave_num.shape[-1]
    dir_coeffs = np.stack([_random_shd_coeffs(max_order, 1, rng) for _ in range(num_pos)], axis=1)
    seq = rng.uniform(-1, 1, size=(seq_len,))
    Phi = sfe._sequence_stft_multiperiod(seq, num_periods)[:num_real_freqs,:]
    return pos, wave_num, dir_coeffs, Phi, seq_len, num_real_freqs

def test_compare_compiled_psi_implementations():
    rng = np.random.default_rng(1234567)
    samplerate = 1000
    c = 343
    seq_len = 300
    num_periods = 2
    num_pos = seq_len * num_periods
    num_freqs = seq_len
    max_order = 1
    pos, wave_num, dir_coeffs, Phi, seq_len, num_real_freqs = _random_psi_arguments(rng, seq_len, num_periods, max_order, samplerate, c)

    profiler = Profiler()
    profiler.start()
    psi_jax = mm.calculate_psi_compiled(pos, dir_coeffs, wave_num, Phi, seq_len, num_real_freqs)
    profiler.stop()
    profiler.print()

    profiler = Profiler()
    profiler.start()
    psi_jax = mm.calculate_psi_compiled(pos, dir_coeffs, wave_num, Phi, seq_len, num_real_freqs)
    profiler.stop()
    profiler.print()

    profiler = Profiler()
    profiler.start()
    psi_numpy = mm.calculate_psi(pos, dir_coeffs, wave_num, Phi, seq_len, num_real_freqs)
    profiler.stop()
    profiler.print()

    assert np.allclose(psi_numpy, psi_jax)


def test_compare_compiled_inf_dimensional_shd_dynamic_implementations():
    rng = np.random.default_rng(1234567)
    samplerate = 1000
    c = 343
    seq_len = 6
    num_periods = 2
    num_pos = seq_len * num_periods
    num_eval = 20
    num_freqs = seq_len
    max_order = 1
    p, pos, pos_eval, seq, reg_param, dir_coeffs = _random_mm_arguments(rng, seq_len, num_periods, max_order, num_eval)

    profiler = Profiler()
    profiler.start()
    p_eval_jax = mm.inf_dimensional_shd_dynamic_compiled(p, pos, pos_eval, seq, samplerate, c, reg_param, dir_coeffs)
    profiler.stop()
    profiler.print()

    profiler = Profiler()
    profiler.start()
    p_eval_numpy = mm.inf_dimensional_shd_dynamic(p, pos, pos_eval, seq, samplerate, c, reg_param, dir_coeffs)
    profiler.stop()
    profiler.print()
    
    assert np.allclose(p_eval_numpy, p_eval_jax)


def test_estimate_from_regressor_is_identical_to_original_estimate():
    rng = np.random.default_rng(1234567)
    seq_len = 20
    num_periods = 2
    num_eval = 15
    max_order = 1
    samplerate = 1000
    c = 343
    p, pos, pos_eval, seq, reg_param, dir_coeffs = _random_mm_arguments(rng, seq_len, num_periods, max_order, num_eval)
    wave_num = ft.get_real_wavenum(seq_len, samplerate, c)

    p_eval, regressor, psi = mm.inf_dimensional_shd_dynamic(p, pos, pos_eval, seq, samplerate, c, reg_param, dir_coeffs, verbose=True)
    p_eval2 = mm.estimate_from_regressor(regressor, pos, pos_eval, wave_num, dir_coeffs)
    assert np.allclose(p_eval, p_eval2)


def test_estimate_from_regressor_compiled_and_numpy_version_are_identical():
    rng = np.random.default_rng(1234567)
    seq_len = 20
    num_periods = 2
    num_eval = 15
    max_order = 1
    samplerate = 1000
    c = 343
    p, pos, pos_eval, seq, reg_param, dir_coeffs = _random_mm_arguments(rng, seq_len, num_periods, max_order, num_eval)
    wave_num = ft.get_real_wavenum(seq_len, samplerate, c)

    p_eval_orig, regressor, psi = mm.inf_dimensional_shd_dynamic(p, pos, pos_eval, seq, samplerate, c, reg_param, dir_coeffs, verbose=True)

    p_eval_jax = mm.estimate_from_regressor_compiled(regressor, pos, pos_eval, wave_num, dir_coeffs)
    p_eval_numpy = mm.estimate_from_regressor(regressor, pos, pos_eval, wave_num, dir_coeffs)
    assert np.allclose(p_eval_jax, p_eval_numpy)




def test_spherical_jn_against_scipy():
    rng = np.random.default_rng(1234567)
    num_points = 10000
    max_val = 200
    order = 2

    n = order * np.ones(num_points, dtype=np.int32)#np.arange(10, dtype=np.int32)
    x = rng.uniform(0, max_val, size=(num_points,))
    sph_jn = mm.spherical_jn(n, x)
    scipy_jn = special.spherical_jn(n, x)

    fig, axes = plt.subplots(2, 1)
    axes[0].plot(x, sph_jn, label="Jax", marker="x", linestyle="")
    axes[0].plot(x, scipy_jn, label="Scipy", marker="x", linestyle="")
    axes[1].plot(x, 20 * np.log10(np.abs(sph_jn - scipy_jn)), label="Difference (dB)", marker="x", linestyle="")
    for ax in axes:
        ax.legend()
        aspplot.set_basic_plot_look(ax)
    plt.show()
    #assert np.allclose(sph_jn, scipy_jn)


def test_translation_operator_against_jax_compiled_implementation():
    """DONT CHANGE BEFORE FIX
    Currently recreates the problem of jax implemententation not matching nunmpy implementation. Seems to 
    be numerically unstable for small inputs, but very non-obvious how to recreate. 
    """
    
    rng = np.random.default_rng(21345654)
    #pos = 1e-5 * np.array([[1,-1,0], [1,-1,0], [1,-1,0]]) # #1e-4 * rng.normal(size = (3,3)) 

    # pos =  np.array([[ 9.54350994e-05, -8.09626527e-05,  9.76747952e-05],
    #                 [-1.60965802e-04, -8.47193020e-06,  1.21416007e-04],
    #                 [-3.36724737e-05, -7.74186296e-06, -1.45218513e-04]])
    pos = np.array([[1e-4, -8e-5, 1e-4]])


    freq = np.array([100, 100])#rng.uniform(100, 1000, size=(2,))
    wave_num = 2 * np.pi * freq / 343
    max_order_input = 1
    max_order_output = 1

    gaunt = mm._calculate_gaunt_set(max_order_input, max_order_output)
    T_jax = mm.translation_operator(pos, wave_num, gaunt)
    T = sph.translation_operator(pos, wave_num, max_order_input, max_order_output)

    T_jax = T_jax[0,0,:,:]
    T = T[0,0,:,:]
    # UNCOMMENT TO VIEW DIFFERENCES
    fig, axes = plt.subplots(1, 3, figsize=(14,4))
    axes[0].set_title("Numpy")
    clr = axes[0].matshow(np.squeeze(np.abs(T)))
    plt.colorbar(clr, ax = axes[0])
    clr = axes[1].matshow(np.squeeze(np.real(T)))
    plt.colorbar(clr, ax = axes[1])
    clr = axes[2].matshow(np.squeeze(np.imag(T)))
    plt.colorbar(clr, ax = axes[2])

    fig, axes = plt.subplots(1, 3, figsize=(14,4))
    axes[0].set_title("Jax")
    clr = axes[0].matshow(np.squeeze(np.abs(T_jax)))
    plt.colorbar(clr, ax = axes[0])
    clr = axes[1].matshow(np.squeeze(np.real(T_jax)))
    plt.colorbar(clr, ax = axes[1])
    clr = axes[2].matshow(np.squeeze(np.imag(T_jax)))
    plt.colorbar(clr, ax = axes[2])

    fig, axes = plt.subplots(1, 3, figsize=(14,4))
    axes[0].set_title("Difference")
    clr = axes[0].matshow(np.squeeze(np.abs(T_jax-T)))
    plt.colorbar(clr, ax = axes[0])
    clr = axes[1].matshow(np.squeeze(np.real(T_jax-T)))
    plt.colorbar(clr, ax = axes[1])
    clr = axes[2].matshow(np.squeeze(np.imag(T_jax-T)))
    plt.colorbar(clr, ax = axes[2])
    plt.show()
    assert np.allclose(T, T_jax)




def test_find_numerical_problems_in_jax_translation_operator():
    rng = np.random.default_rng(21345654)
    num_points = 300
    num_scales = 100


    test_dirs = mc.uniform_random_on_sphere(num_points, rng)
    scalings = np.logspace(-10, 3, num_scales)
    pos = test_dirs[None,:,:] * scalings[:,None,None]
    #pos = pos.reshape(-1, 3)

    freq = np.array([100])#rng.uniform(100, 1000, size=(2,))
    wave_num = 2 * np.pi * freq / 343
    max_order_input = 1
    max_order_output = 1

    gaunt = mm._calculate_gaunt_set(max_order_input, max_order_output)

    mse = []
    maxse = []
    for i, s in enumerate(scalings):
        T_jax = mm.translation_operator(pos[i,:,:], wave_num, gaunt)
        T = sph.translation_operator(pos[i,:,:], wave_num, max_order_input, max_order_output)
        mse_s = np.mean(np.abs(T_jax - T)**2, axis=(-1, -2))
        mse.append(np.mean(mse_s))
        maxse.append(np.max(mse_s))
        #mse.append(np.mean(np.abs(T_jax - T)**2) / np.mean(np.abs(T)**2))
        #maxse.append(np.max(np.abs(T_jax - T)**2))

    i = np.argmax(maxse)
    max_bessel_arg = wave_num * scalings[i]

    fig, ax = plt.subplots(1, 1)
    ax.plot(wave_num*scalings, mse, label="MSE")
    ax.plot(wave_num*scalings, maxse, label="MaxSE")
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.set_xlabel("Bessel argument")
    ax.set_ylabel("Error")
    ax.set_title(f"Max error at {max_bessel_arg}")
    ax.legend()
    plt.show()







def test_spherical_bessel_function_numerical_stability():
    rng = np.random.default_rng(21345654)
    num_points = 3000

    arg = np.logspace(-10, 3, num_points)
    max_order = 2

    jn_jax = []
    jn_scipy = []
    for i in range(max_order+1):
        jn_scipy.append(special.spherical_jn((i * np.ones_like(arg)).astype(int), arg))
        jn_jax.append(mm.spherical_jn((i * np.ones_like(arg)).astype(int), arg))
    

    error = [jn_j - jn_s for jn_j, jn_s in zip(jn_jax, jn_scipy)]

    fig, axes = plt.subplots(max_order+1, 2, sharex=True, figsize=(16, 9))

    for i in range(max_order+1):
        axes[i][0].plot(arg, jn_jax[i], label="Jax")
        axes[i][0].plot(arg, jn_scipy[i], label="Scipy")
        axes[i][1].plot(arg, np.abs(error[i]), label="Error")
    for ax_row in axes:
        for ax in ax_row:
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_ylabel("Amplitude")
            aspplot.set_basic_plot_look(ax)
            ax.legend()
    axes[0,0].set_title("Function values")
    axes[0,1].set_title("Error")
    axes[-1,0].set_xlabel("Bessel argument")
    axes[-1,1].set_xlabel("Bessel argument")
    plt.show()



def test_bessel_function_integer_order_numerical_stability():
    rng = np.random.default_rng(21345654)
    num_points = 3000

    arg = np.logspace(-10, -3, num_points)
    max_order = 2

    jn_jax = []
    jn_scipy = []
    jn_jax.append(mm.j0(arg))
    jn_jax.append(mm.j1(arg))
    #or i in range(2, max_order+1):
    #   jn_jax.append(mm.jv(i, arg))

    #jn_jax.append(arg**2 / 15)


    for i in range(max_order+1):
        jn_scipy.append(special.jv((i * np.ones_like(arg)).astype(int), arg))
        #jn_scipy.append(special.jv((i * np.ones_like(arg)).astype(int), arg))

    error = [jn_j - jn_s for jn_j, jn_s in zip(jn_jax, jn_scipy)]

    num_orders = len(jn_jax)
    fig, axes = plt.subplots(num_orders, 2, sharex=True, figsize=(16, 9))

    for i in range(num_orders):
        axes[i][0].plot(arg, jn_jax[i], label="Jax")
        axes[i][0].plot(arg, jn_scipy[i], label="Scipy")
        axes[i][1].plot(arg, np.abs(error[i]), label="Error")
    for ax_row in axes:
        for ax in ax_row:
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_ylabel("Amplitude")
            aspplot.set_basic_plot_look(ax)
            ax.legend()
    axes[0,0].set_title("Function values")
    axes[0,1].set_title("Error")
    axes[-1,0].set_xlabel("Bessel argument")
    axes[-1,1].set_xlabel("Bessel argument")
    plt.show()



def test_bessel_function_half_integer_order_numerical_stability():
    rng = np.random.default_rng(21345654)
    num_points = 3000

    arg = np.logspace(-10, 6, num_points)
    max_order = 2

    jn_jax = []
    jn_scipy = []

    #for i in range(max_order+1):
    #    jn_scipy.append(special.jv(((0.5+i) * np.ones_like(arg)).astype(int), arg))
        #jn_scipy.append(special.jv((i * np.ones_like(arg)).astype(int), arg))

    jn_jax.append(mm.jneghalf(arg))
    jn_jax.append(mm.jposhalf(arg))
    jn_jax.append(mm.jvplushalf(2, arg))

    jn_scipy.append(special.jv(-0.5, arg))
    jn_scipy.append(special.jv(0.5, arg))
    jn_scipy.append(special.jv(2.5, arg))

    error = [jn_j - jn_s for jn_j, jn_s in zip(jn_jax, jn_scipy)]

    num_orders = len(jn_jax)
    fig, axes = plt.subplots(num_orders, 2, sharex=True, figsize=(16, 9))

    for i in range(num_orders):
        axes[i][0].plot(arg, jn_jax[i], label="Jax")
        axes[i][0].plot(arg, jn_scipy[i], label="Scipy")
        axes[i][1].plot(arg, np.abs(error[i]), label="Error")
    for ax_row in axes:
        for ax in ax_row:
            ax.set_xscale("log")
            ax.set_yscale("log")
            ax.set_ylabel("Amplitude")
            aspplot.set_basic_plot_look(ax)
            ax.legend()
    axes[0,0].set_title("Function values")
    axes[0,1].set_title("Error")
    axes[-1,0].set_xlabel("Bessel argument")
    axes[-1,1].set_xlabel("Bessel argument")
    plt.show()
















